/*
    These two functions used for rough reflections are based on zombye's spectrum shaders
    https://github.com/zombye/spectrum
*/

mat3 getRotationMat(vec3 x, vec3 y) {
	float cosine = dot(x, y);
	vec3 axis = cross(y, x);

	float tmp = 1.0 / dot(axis, axis);
	      tmp = tmp - tmp * cosine;
	vec3 tmpv = axis * tmp;

	return mat3(
		axis.x * tmpv.x + cosine, axis.x * tmpv.y - axis.z, axis.x * tmpv.z + axis.y,
		axis.y * tmpv.x + axis.z, axis.y * tmpv.y + cosine, axis.y * tmpv.z - axis.x,
		axis.z * tmpv.x - axis.y, axis.z * tmpv.y + axis.x, axis.z * tmpv.z + cosine
	);
}

#define GGX_TAILCLAMP 0.25

vec3 ggxFacetDist(vec3 viewDir, float roughness, vec2 xy) {
	/*
        GGX VNDF sampling
        http://www.jcgt.org/published/0007/04/01/
    */
    roughness   = max(roughness, 0.001);
    xy.x        = clamp(xy.x * GGX_TAILCLAMP, 0.001, GGX_TAILCLAMP);

    viewDir     = normalize(vec3(roughness * viewDir.xy, viewDir.z));

    float clsq  = dot(viewDir.xy, viewDir.xy);
    vec3 T1     = vec3(clsq > 0.0 ? vec2(-viewDir.y, viewDir.x) * inversesqrt(clsq) : vec2(1.0, 0.0), 0.0);
    vec3 T2     = vec3(-T1.y * viewDir.z, viewDir.z * T1.x, viewDir.x * T1.y - T1.x * viewDir.y);

	float r     = sqrt(xy.x);
	float phi   = tau * xy.y;
	float t1    = r * cos(phi);
	float a     = saturate(1.0 - t1 * t1);
	float t2    = mix(sqrt(a), r * sin(phi), 0.5 + 0.5 * viewDir.z);

	vec3 normalH = t1 * T1 + t2 * T2 + sqrt(saturate(a - t2 * t2)) * viewDir;

	return normalize(vec3(roughness * normalH.xy, normalH.z));
}

float ggxWeight(vec3 viewDir, vec3 microfacet, float roughness, float noise) {
        noise           = clamp(noise * GGX_TAILCLAMP, 0.001, GGX_TAILCLAMP);
    float alpha2        = sqr(roughness);
    float cosTheta2     = (1.0 - noise) / (noise * (alpha2 - 1.0) + 1.0);

    float den           = (alpha2 - 1.0) * cosTheta2 + 1.0;
    float D             = alpha2 / (pi * sqr(den));
    float p             = D * sqrt(cosTheta2) / (4.0 * dot(microfacet, viewDir));

    return rcp((0.5 / pi) / (p + 1e-8));
}  

/*
    Screenspace Trace
*/

#ifdef voxelPT_ReflectionsEnabled
    #define SSR_DistanceThreshold 2.0
#else
    #define SSR_DistanceThreshold 128.0
#endif

vec3 screenspaceRT(vec3 position, vec3 direction, float noise) {
    const uint maxSteps     = 16;

  	float rayLength = ((position.z + direction.z * far * sqrt3) > -near) ?
                      (-near - position.z) / direction.z : far * sqrt3;

    vec3 screenPosition     = viewToScreenSpace(position);
    vec3 endPosition        = position + direction * rayLength;
    vec3 endScreenPosition  = viewToScreenSpace(endPosition);

    vec3 screenDirection    = normalize(endScreenPosition - screenPosition);
        screenDirection.xy  = normalize(screenDirection.xy);

    vec3 maxLength          = (step(0.0, screenDirection) - screenPosition) / screenDirection;
    float stepMult          = minOf(maxLength);
    vec3 screenVector       = screenDirection * stepMult / float(maxSteps);

    vec3 screenPos          = screenPosition + screenDirection * maxOf(pixelSize * pi);

    if (saturate(screenPos.xy) == screenPos.xy) {
        float depthSample   = texelFetch(depthtex0, ivec2(screenPos.xy * viewSize), 0).x;
        float linearSample  = depthLinear(depthSample);
        float currentDepth  = depthLinear(screenPos.z);

        if (linearSample < currentDepth) {
            float dist      = abs(linearSample - currentDepth) / clamp(currentDepth, 0.25 / far, SSR_DistanceThreshold / far);
            if (dist <= 0.25) return vec3(screenPos.xy, depthSample);
        }
    }

        screenPos          += screenVector * noise;

    for (uint i = 0; i < maxSteps; ++i) {
        if (saturate(screenPos.xy) != screenPos.xy) break;

        float depthSample   = texelFetch(depthtex0, ivec2(screenPos.xy * viewSize), 0).x;
        float linearSample  = depthLinear(depthSample);
        float currentDepth  = depthLinear(screenPos.z);

        if (linearSample < currentDepth) {
            float dist      = abs(linearSample - currentDepth) / clamp(currentDepth, 0.25 / far, SSR_DistanceThreshold / far);
            if (dist <= 0.25) return vec3(screenPos.xy, depthSample);
        }

        screenPos      += screenVector;
    }

    return vec3(1.1);
}
vec3 screenspaceRT_LR(vec3 position, vec3 direction, float noise) {
    const uint maxSteps     = 8;

  	float rayLength = ((position.z + direction.z * far * sqrt3) > -near) ?
                      (-near - position.z) / direction.z : far * sqrt3;

    vec3 screenPosition     = viewToScreenSpace(position);
    vec3 endPosition        = position + direction * rayLength;
    vec3 endScreenPosition  = viewToScreenSpace(endPosition);

    vec3 screenDirection    = normalize(endScreenPosition - screenPosition);
        screenDirection.xy  = normalize(screenDirection.xy);

    vec3 maxLength          = (step(0.0, screenDirection) - screenPosition) / screenDirection;
    float stepMult          = min(minOf(maxLength), 0.5);
    vec3 screenVector       = screenDirection * stepMult / float(maxSteps);

    vec3 screenPos          = screenPosition;

        screenPos          += screenVector * noise;

    for (uint i = 0; i < maxSteps; ++i) {
        if (saturate(screenPos.xy) != screenPos.xy) break;

        float depthSample   = texelFetch(depthtex0, ivec2(screenPos.xy * viewSize), 0).x;
        float linearSample  = depthLinear(depthSample);
        float currentDepth  = depthLinear(screenPos.z);

        if (linearSample < currentDepth) {
            float dist      = abs(linearSample - currentDepth) / clamp(currentDepth, 0.25 / far, SSR_DistanceThreshold / far);
            if (dist <= 0.25) return vec3(screenPos.xy, depthSample);
        }

        screenPos      += screenVector;
    }

    return vec3(1.1);
}


/*
    Voxel Trace
*/

#if DIM != -1
    vec3 getShadowmapPos(vec3 scenePos, const float bias) {  //shadow 2d
        vec3 pos    = scenePos + vec3(bias) * lightDir;
        float a     = length(pos);
            pos     = transMAD(shadowModelView, pos);
            pos     = projMAD(shadowProjection, pos);
            pos.z  *= 0.2;
            pos.z  -= 0.0012*(saturate(a/256.0));

            pos.xy  = shadowmapWarp(pos.xy);

            pos.x   = pos.x * 0.5 + 0.5;
            pos.xy  = pos.yx;

        return pos*0.5+0.5;
    }
    vec3 getDirectLight(vec3 scenePos, vec3 sceneNormal) {
        float sDotN     = dot(lightDir, sceneNormal);
        if (sDotN <= 0.0) return vec3(0.0);

        float shadow    = 1.0;

        #ifndef DIM
            #ifdef cloudShadowsEnabled
                shadow *= readCloudShadowmap(colortex4, scenePos);

                if (shadow < 1e-4) return vec3(0.0);
            #endif
        #endif

        vec3 shadowPos  = getShadowmapPos(scenePos, 0.08);
            shadow  *= texture(shadowtex1, shadowPos);

        return texture(shadowcolor0, shadowPos.xy).rgb * (shadow * saturate(sDotN));
    }
#endif

#define atlasSampler colortex8
#define atlasSamplerSPEC colortex7

int atlasMipmapLevel = textureQueryLevels(atlasSampler) - 1;
ivec2 atlasSize      = ivec2(textureSize(atlasSampler, 0));

#define RPASS

#include "/lib/voxel/store.glsl"
#include "/lib/voxel/intersect.glsl"
#include "/lib/voxel/trace.glsl"
#include "/lib/mat/emission.glsl"

#include "/lib/brdf/hammon.glsl"

struct Material{
    vec3 normal;
    vec3 albedo;
    vec3 emissive;

    int id;
};

Material hitMaterial(vec3 normal, vec3 albedo, int voxelID) {
    albedo  = linearToAP1Albedo(toLinear(albedo));
    
    return Material(normal, albedo, getEmission(voxelID, albedo), voxelID);
}

float getLabEmission(ivec2 uv) {
    /*
    float x     = texelFetch(atlasSampler_s, uv, 0).a * 255;
    return x <= (254.5) ? linStep(x, 0.5, 254.5) : 0.0;*/
    return 0.0;
}

vec3 getEmission(int ID, vec3 albedo, ivec2 atlasUV) {
    albedo  = max(albedo, 1e-8);
    vec3 albedoIn   = albedo;

    #if labEmissionMode != 0
    float textureEmission = getLabEmission(atlasUV); 
    #endif

    #if labEmissionMode != 2
        float lum   = getLuma(albedo);

        float albedoLum = mix(avgOf(albedo), maxOf(albedo), 0.71);
            albedoLum   = saturate(albedoLum * sqrt2);

        float emitterLum = saturate(mix(sqr(albedoLum), sqrt(maxOf(albedo)), albedoLum) * sqrt2);

        #if labEmissionMode == 1
            emitterLum  = max(emitterLum, textureEmission);
        #endif

        albedo  = mix(sqr(normalize(albedo)), normalize(albedo), sqrt(emitterLum)) * emitterLum;
    #else       
        albedo   *= textureEmission;
        float emitterLum = textureEmission;
    #endif

    switch (ID) {
        case 1: return albedo * 200;
        case 2: return albedo * 160;
        case 3: return albedo * 80;
        case 4: return albedo * 10;
        case 10: return LIGHT_TORCH * 120;
        case 11: return LIGHT_REDTORCH * 35;
        case 12: return LIGHT_SOUL * 60;
        case 13: return LIGHT_FIRE * 80;
        case 20: return LIGHT_ENDROD * 100;
        case 21: return LIGHT_AMETHYST * 20;
        case 22: return LIGHT_AMETHYST * 5;
    }

    #if labEmissionMode == 0
        return vec3(0);
    #elif labEmissionMode == 1
        return albedoIn * textureEmission * 80;
    #elif labEmissionMode == 2
        return albedo * 80;
    #endif
}

Material hitMaterial(vec3 normal, vec3 albedo, int voxelID, ivec2 atlasUV) {
    albedo  = linearToAP1Albedo(toLinear(albedo));
    
    return Material(normal, albedo, getEmission(voxelID, albedo, atlasUV), voxelID);
}

#define useMipBounceAlbedo

Material getVoxelMat(vec3 pos, vec3 normal, voxelRead voxelData, in int bounceIndex) {
    vec2 uv     = voxelUV(pos, normal, voxelData);

    #ifdef useMipBounceAlbedo
        int lod     = min(min(bounceIndex, atlasMipmapLevel), 1);
        vec4 albedo = texelFetch(atlasSampler, ivec2(uv * atlasSize / max(lod * 2, 1)), lod);
    #else
        vec4 albedo = texelFetch(atlasSampler, ivec2(uv * atlasSize), 0);
    #endif

    return hitMaterial(normal, albedo.rgb * voxelData.tint, voxelData.ID.y, ivec2(uv * atlasSize));
}
Material getVoxelMat(vec3 pos, vec3 normal, voxelRead voxelData, in int bounceIndex, inout materialProperties material) {
    vec2 uv     = voxelUV(pos, normal, voxelData);

    #ifdef useMipBounceAlbedo
        int lod     = min(min(bounceIndex, atlasMipmapLevel), 1);
        vec4 albedo = texelFetch(atlasSampler, ivec2(uv * atlasSize / max(lod * 2, 1)), lod);
        vec2 matData = texelFetch(atlasSamplerSPEC, ivec2(uv * atlasSize / max(lod * 2, 1)), lod).xy;
    #else
        vec4 albedo = texelFetch(atlasSampler, ivec2(uv * atlasSize), 0);
        vec2 matData = texelFetch(atlasSamplerSPEC, ivec2(uv * atlasSize), 0).xy;
    #endif

    material    = decodeLabBasic(matData);

    return hitMaterial(normal, albedo.rgb * voxelData.tint, voxelData.ID.y, ivec2(uv * atlasSize));
}

#define HASHSCALE3 vec3(.1031, .1030, .0973)
vec2 hash22(vec2 p) {
	vec3 p3 = fract(vec3(p.xyx) * HASHSCALE3);
    p3 += dot(p3, p3.yzx + 19.19);
    return fract((p3.xx + p3.yz) * p3.zy);
}
vec3 genUnitVector(vec2 p) {
    p.x *= tau; p.y = p.y * 2.0 - 1.0;
    return vec3(sincos(p.x) * sqrt(1.0 - p.y * p.y), p.y);
}
vec3 GenerateCosineVectorSafe(vec3 vector, vec2 xy) {
	// Apparently this is actually this simple.
	// http://www.amietia.com/lambertnotangent.html
	// (cosine lobe around vector = lambertian BRDF)
	// This one deals with ther rare case where cosineVector == (0,0,0)
	// Can just normalize it instead if you are sure that won't be a problem
	vec3 cosineVector = vector + genUnitVector(xy);
	float lenSq = dot(cosineVector, cosineVector);
	return lenSq > 0.0 ? cosineVector * inversesqrt(lenSq) : vector;
}

uniform int frame255;

vec4 reflectionTrace(vec3 viewPosition, vec3 viewNormal, vec3 viewDirection) {
    vec3 scenePosition  = viewToSceneSpace(viewPosition);
    vec3 sceneNormal    = mat3(gbufferModelViewInverse) * viewNormal;
    vec3 direction      = mat3(gbufferModelViewInverse) * viewDirection;

    vec3 voxelPos       = sceneToVoxelSpace(scenePosition);
    vec3 voxelIndex     = floor(voxelPos - sceneNormal / 128.0);
    vec3 startIndex     = voxelIndex;
    vec3 cameraPosition = sceneToVoxelSpace(gbufferModelViewInverse[3].xyz);

    if (outsideVoxelVolume(voxelIndex)) return vec4(0);

    vec3 rayColor       = vec3(0);
    vec3 contribution   = vec3(1);

    const float a1 = 1.0 / rho;
    const float a2 = a1 * a1;

    voxelRead voxelData;
    Material hitMat;

    #if DIM != -1
    vec3 directCol      = (sunAngle<0.5 ? lightColor[0] : lightColor[2]) * pi;
    #endif

    vec3 hitNormal      = sceneNormal;

    bool hit            = marchBVH(voxelPos, ivec3(voxelIndex), direction, voxelPos, voxelIndex, hitNormal, voxelData, contribution);

    if (hit) {
        hitMat      = getVoxelMat(voxelPos, hitNormal, voxelData, 0);
        vec3 radiance = hitMat.emissive * rpi;
            radiance += sampleCacheHQ(voxelToSceneSpace(voxelPos), hitNormal, getFace(hitNormal)) * hitMat.albedo;

        #if DIM != -1
            radiance += directCol * getDirectLight(voxelToSceneSpace(voxelPos), hitNormal) * hitMat.albedo * rpi;
        #endif

        rayColor   += contribution * radiance;
        
        #ifdef voxelPT_ReflectionLightBounce
        vec2 quasirandomCurr = 0.5 + fract(vec2(a1, a2));
        vec2 noiseCurr  = hash22(gl_FragCoord.xy + frame255);

        noiseCurr      += hash22(vec2(gl_FragCoord.xy + vec2(cos(quasirandomCurr.x), sin(quasirandomCurr.y))));
        vec3 oldDirection   = direction;

        vec2 vectorXY   = fract(sqrt(2.0) * quasirandomCurr + noiseCurr);

        direction       = GenerateCosineVectorSafe(hitNormal, vectorXY);

        vec3 brdf = hitMat.albedo;
            brdf *= diffuseHammon(hitNormal, -oldDirection, direction, 1.0) / saturate(dot(direction, hitNormal)/pi);

        contribution *= saturate(brdf);

        hit = marchBVH(voxelPos, ivec3(voxelIndex), direction, voxelPos, voxelIndex, hitNormal, voxelData, contribution);
        if (hit) {
            //rayColor   += contribution * (hitMat.emissive + (directCol * getDirectLight(voxelToSceneSpace(voxelPos), hitNormal)) * hitMat.albedo);
            rayColor   += contribution * (cacheRead(getCachePixel(voxelToSceneSpace(voxelPos) - hitNormal * 0.05, getFace(hitNormal))).rgb*rpi * hitMat.albedo);
        }
        #endif

    return vec4(rayColor, 1.0);
    }

    return vec4(0);
}

vec4 reflectionMultibounce(vec3 viewPosition, vec3 viewNormal, vec3 viewDirection, vec3 blueNoise, materialProperties material, float occlusion) {
    vec3 scenePosition  = viewToSceneSpace(viewPosition);
    vec3 sceneNormal    = mat3(gbufferModelViewInverse) * viewNormal;
    vec3 direction      = mat3(gbufferModelViewInverse) * viewDirection;

    vec3 voxelPos       = sceneToVoxelSpace(scenePosition);
    vec3 voxelIndex     = floor(voxelPos - sceneNormal / 128.0);
    vec3 startIndex     = voxelIndex;
    vec3 cameraPosition = sceneToVoxelSpace(gbufferModelViewInverse[3].xyz);

    if (outsideVoxelVolume(voxelIndex)) return vec4(0);
    if (material.roughness > roughnessThreshold) return vec4(0);

    vec3 rayColor       = vec3(0);
    vec3 contribution   = vec3(1);

    const float a1 = 1.0 / rho;
    const float a2 = a1 * a1;

    voxelRead voxelData;
    Material hitMat;

    #if DIM != -1
    vec3 directCol      = (sunAngle<0.5 ? lightColor[0] : lightColor[2]) * pi * lightFlip;
    #endif

    vec3 hitNormal      = sceneNormal;

    bool anyHit         = false;
    bool doSSR          = true;

    float skyAlpha      = 0.0;

    for (uint i = 0; i < SPECULAR_BOUNCES; ++i) {
        vec3 previousDirection  = direction;
        bool hit        = false;
        if (dot(direction, hitNormal) > 0.0) hitNormal = -hitNormal;

        if (material.roughness > roughnessThreshold) break;

        #ifndef resourcepackReflectionsEnabled
        if (i > 0) break;
        #endif

        #ifdef screenspaceReflectionsEnabled
        if (doSSR) {
            vec3 reflectedPos   = vec3(1.1);
            vec3 viewPosition   = viewToSceneSpace(voxelToSceneSpace(voxelPos), gbufferModelView);
            vec3 viewDirection  = mat3(gbufferModelView) * direction;
            if (material.roughness < 0.3) reflectedPos = screenspaceRT(viewPosition, viewDirection, blueNoise.z);
            else reflectedPos = screenspaceRT_LR(viewPosition, viewDirection, blueNoise.z);

            hit     = reflectedPos.z < 1.0;

            anyHit  = hit || anyHit;

            if (hit) {
                ivec2 UV    = ivec2(reflectedPos.xy * viewSize);

                material    = decodeLabBasic(unpack2x8(texelFetch(colortex2, UV, 0).x));

                if (i == SPECULAR_BOUNCES-1) {
                    material.conductor = false;
                }

                mat2x3 lightingData = unpackDoubleRGBE8(texelFetch(colortex5, UV, 0));
                #if DIM != -1
                    lightingData[0] = min(lightingData[0] * pi, directCol);
                #endif

                hitNormal   = decodeNormal(texelFetch(colortex1, UV, 0).xy);

                voxelPos    = sceneToVoxelSpace(viewToSceneSpace(screenToViewSpace(reflectedPos)));
                voxelIndex  = floor(voxelPos - hitNormal / 128.0);

                /*
                    Reflect Ray
                */
                blueNoise       = fract(blueNoise + torroidalShift.xyz);

                mat3 rot        = getRotationMat(vec3(0, 0, 1), mat3(gbufferModelView) * hitNormal);
                vec3 tangentV   = (mat3(gbufferModelView) * direction) * rot;

                vec3 microfacet = material.roughness > 1e-6 ? mat3(gbufferModelViewInverse) * (rot * ggxFacetDist(-tangentV, (material.roughness), blueNoise.xy)) : hitNormal;
                direction       = reflect(direction, microfacet);
                if (dot(direction, hitNormal) < 0.0) direction = -direction;

                #if DIM != -1
                    vec3 directSpecular = lightingData[0] * specularTrowbridgeReitzGGX(-previousDirection, lightDir, hitNormal, material, hitMat.albedo);
                #else
                    const vec3 directSpecular = vec3(0);
                #endif

                if (material.conductor) {
                    rayColor       += directSpecular * contribution;
                } else {
                    rayColor       += (texelFetch(colortex0, UV, 0).rgb + directSpecular) * contribution;

                    contribution   *= lightingData[1];
                }

                contribution *= BRDFfresnelAlbedoTint(-previousDirection, hitNormal, material, saturate(lightingData[1]));

                #ifdef voxelPT_ReflectionsEnabled
                if (i > 2) doSSR = false;
                #endif

                continue;
            }

            if (!hit) {
                doSSR = false;
            }
        }
        #endif

        #ifdef voxelPT_ReflectionsEnabled

        if (!hit) hit = marchBVH(voxelPos, ivec3(voxelIndex), direction, voxelPos, voxelIndex, hitNormal, voxelData, contribution);

        anyHit  = hit || anyHit;

        if (hit) {
            hitMat      = getVoxelMat(voxelPos, hitNormal, voxelData, 0, material);
            
            #ifndef roughReflectionsEnabled
            material.roughness = 0;
            #endif

            vec3 BRDF       = specularTrowbridgeReitzGGX(previousDirection, direction, hitNormal, material, hitMat.albedo);

            /*
                Reflect Ray
            */
            blueNoise       = fract(blueNoise + torroidalShift.xyz);

            mat3 rot        = getRotationMat(vec3(0, 0, 1), mat3(gbufferModelView) * hitNormal);
            vec3 tangentV   = (mat3(gbufferModelView) * direction) * rot;

            vec3 microfacet = material.roughness > 1e-6 ? mat3(gbufferModelViewInverse) * (rot * ggxFacetDist(-tangentV, (material.roughness), blueNoise.xy)) : hitNormal;
            direction       = reflect(direction, microfacet);
            if (dot(direction, hitNormal) < 0.0) direction = -direction;

            #if DIM != -1
                vec3 directLight    = directCol * getDirectLight(voxelToSceneSpace(voxelPos), hitNormal);
                vec3 directSpecular = directLight * specularTrowbridgeReitzGGX(-previousDirection, lightDir, hitNormal, material, hitMat.albedo);
            #else
                const vec3 directSpecular = vec3(0);
            #endif

            vec3 cacheRadiance  = sampleCacheHQ(voxelToSceneSpace(voxelPos), hitNormal, getFace(hitNormal));

            if (i == SPECULAR_BOUNCES-1) {
                material.conductor = false;
            }

            if (material.conductor) {                
                rayColor       += (directSpecular + hitMat.emissive * BRDF) * contribution;
            } else {
                vec3 radiance = hitMat.emissive * rpi;
                    radiance += cacheRadiance * hitMat.albedo;

                #if DIM != -1
                    radiance += directLight * hitMat.albedo * rpi + directSpecular;
                #endif

                rayColor   += contribution * radiance;
                contribution   *= hitMat.albedo;
            }

            contribution   *= BRDFfresnelAlbedoTint(-previousDirection, hitNormal, material, hitMat.albedo);

            if (maxOf(contribution) < 1e-4) break;

            /*
            #ifdef voxelPT_ReflectionLightBounce
                vec2 quasirandomCurr = 0.5 + fract(vec2(a1, a2));
                vec2 noiseCurr  = hash22(gl_FragCoord.xy + frame255);

                noiseCurr      += hash22(vec2(gl_FragCoord.xy + vec2(cos(quasirandomCurr.x), sin(quasirandomCurr.y))));
                vec3 oldDirection   = previousDirection;

                vec2 vectorXY   = fract(sqrt(2.0) * quasirandomCurr + noiseCurr);

                vec3 LVDirection = GenerateCosineVectorSafe(hitNormal, vectorXY);

                vec3 brdf = hitMat.albedo;
                    brdf *= diffuseHammon(hitNormal, -oldDirection, LVDirection, 1.0) / saturate(dot(LVDirection, hitNormal)/pi);

                vec3 LVContribution = contribution * saturate(brdf);

                vec3 LVIndex    = voxelIndex;
                vec3 LVPos      = voxelPos;
                vec3 LVHitNormal = hitNormal;
                voxelRead LVoxel    = voxelData;


                bool LHit = marchBVH(LVPos, ivec3(LVIndex), LVDirection, LVPos, LVIndex, LVHitNormal, LVoxel, LVContribution);
                if (LHit) {
                    rayColor   += LVContribution * (cacheRead(getCachePixel(voxelToSceneSpace(LVPos) - LVHitNormal * 0.05, getFace(LVHitNormal))).rgb*rpi * hitMat.albedo);
                }
            #endif
            */
        }

        if (!hit) {
            rayColor += texture(colortex4, projectSky(direction, 2)).rgb * contribution;
            skyAlpha = 1.0;
            break;
        }
        #else
        if (!hit) {
            rayColor += texture(colortex4, projectSky(direction, 2)).rgb * contribution;
            skyAlpha = occlusion;
            break;
        }
        #endif

    }

    return vec4(rayColor, max(float(anyHit), skyAlpha));
}

vec3 getRoughFresnel(vec3 viewNormal, vec3 viewDir, vec2 noise, materialProperties material, vec3 albedo) {
    if (material.roughness < 0.001) return BRDFfresnelAlbedoTint(-viewDir, viewNormal, material, albedo);

    mat3 rot        = getRotationMat(vec3(0, 0, 1), viewNormal);
    vec3 tangentV   = viewDir * rot;

    const uint steps    = 3;
    const float rSteps  = 1.0 / float(steps);

    vec3 fresnel        = vec3(0);

    for (uint i = 0; i < steps; ++i) {
        vec3 roughNrm   = rot * ggxFacetDist(-tangentV, material.roughness, noise.xy);
        noise           = fract(noise + torroidalShift.xy);

            fresnel    += BRDFfresnelAlbedoTint(-viewDir, roughNrm, material, albedo);
    }
    fresnel    *= rSteps;

    return fresnel;
}