/*
    Cache voxel structure based on Superscalar by bobcao3
*/

#define CacheSizeMultiplier 1


const int volume_width = 64 * CacheSizeMultiplier;
const int volume_height = 64 * CacheSizeMultiplier;

const int cacheSegmentResolution = 512 * CacheSizeMultiplier;
const int cacheSegmentEdge = cacheSegmentResolution - 1;

const int volume_depth_grid_width = (cacheSegmentResolution / volume_width);
const int volume_depth_grid_height = (cacheSegmentResolution / volume_height);

const int volume_depth = (cacheSegmentResolution / volume_width) * (cacheSegmentResolution / volume_height);

ivec3 sceneToCacheSpace(vec3 world_pos) {
    return ivec3(floor(world_pos + mod(cameraPosition, 1.0))) + ivec3(volume_width, volume_depth, volume_height) / 2;
}
ivec3 sceneToCacheSpaceNoCam(vec3 world_pos) {
    return ivec3(floor(world_pos)) + ivec3(volume_width, volume_depth, volume_height) / 2;
}
vec3 cacheToSceneSpace(ivec3 cachePos) {
    return vec3(cachePos) - vec3(volume_width, volume_depth, volume_height) / 2.0;
}

bool outsideCacheVolume(ivec3 pos) {
    return pos.x < 0 || pos.x >= volume_width || pos.y < 0 || pos.y >= volume_depth || pos.z < 0 || pos.z >= volume_height;
}
bool sceneWithinCacheVolume(vec3 pos) {
    return !outsideCacheVolume(sceneToCacheSpaceNoCam(pos));
}

ivec2 getCachePixel(ivec3 pos) {
    if (pos.x < 0 || pos.x >= volume_width || pos.y < 0 || pos.y >= volume_depth || pos.z < 0 || pos.z >= volume_height) return ivec2(-1);

    ivec2 base = ivec2((pos.y % volume_depth_grid_width) * volume_width, (pos.y / volume_depth_grid_width) * volume_height);
    ivec2 uv = base + pos.xz;

    return uv;
}

vec2 getCacheUV(vec3 pos) {
    if (pos.x < 0 || pos.x >= volume_width || pos.y < 0 || pos.y >= volume_depth || pos.z < 0 || pos.z >= volume_height) return ivec2(-1);

    ivec2 base = ivec2((int(pos.y) % volume_depth_grid_width) * volume_width, (int(pos.y) / volume_depth_grid_width) * volume_height);
    vec2 uv = vec2(base) + pos.xz;

    return uv;
}

ivec3 uvToCacheIndex(ivec2 uv) {
    ivec2 y_grid = uv / ivec2(volume_width, volume_height);
    ivec2 xz_pos = uv % ivec2(volume_width, volume_height);
    ivec3 wpos = ivec3(xz_pos.x, y_grid.x + y_grid.y * volume_depth_grid_width, xz_pos.y);

    return wpos;
}

ivec2 getCachePixel(vec3 scenePos, int face) {
    ivec3 cacheSpace = sceneToCacheSpace(scenePos);
    ivec2 uv        = getCachePixel(cacheSpace);

        uv         += ivec2(float(face & 3) * cacheSegmentResolution, floor(face / 4.0) * cacheSegmentResolution);
    return uv;
}
ivec2 getCachePixelCacheSpace(ivec3 cacheSpace, int face) {
    ivec2 uv        = getCachePixel(cacheSpace);

        uv         += ivec2(float(face & 3) * cacheSegmentResolution, floor(face / 4.0) * cacheSegmentResolution);

    return uv;
}

int getFaceFromUV(ivec2 UV) {
    int face    = int(floor(float(UV.x) / cacheSegmentResolution));
        face   += int(floor(float(UV.y) / cacheSegmentResolution)) * 4;
    return face;
}

/*
    Based on ApolloRT by Snurf
*/

const vec3[6] faceNormals = vec3[6] (vec3(1,0,0), vec3(0,1,0), vec3(0,0,1),
									  vec3(-1,0,0), vec3(0,-1,0), vec3(0,0,-1));


const ivec3[6] faceTangents = ivec3[6] (ivec3(0,0,1), ivec3(1,0,0), ivec3(1,0,0),
									  ivec3(0,0,1), ivec3(1,0,0), ivec3(1,0,0));

const ivec3[6] faceBitangents = ivec3[6] (ivec3(0,1,0), ivec3(0,0,1), ivec3(0,1,0),
									  ivec3(0,1,0), ivec3(0,0,1), ivec3(0,1,0));

const int[6] faceHorizontal = int[6] (2, 0, 0, 2, 0, 0);
const int[6] faceVertical = int[6] (1, 2, 1, 1, 2, 1);

int getFace(vec3 n) {//Thanks Capt Tatsu for this function
  ivec3 i = ivec3(n + sign(n)*0.1);
  int three = int(clamp(-(i.x+i.y+i.z), 0, 1))*3;
  return 2 * abs(i.z) + (abs(i.y) + three);
}

#ifdef useIrradianceCache
vec3 sampleCacheHQ(vec3 scenePos, vec3 geoNormal, int face) {
    vec3 scenePosNew = scenePos - geoNormal * 0.25;

	vec3 a = texelFetch(colortex15, getCachePixel((scenePosNew), face), 0).rgb;
	float luma_a = getLuma(a);
	vec3 b = texelFetch(colortex15, getCachePixel((scenePosNew) + faceTangents[face], face), 0).rgb;
	float luma_b = getLuma(b);
	vec3 c = texelFetch(colortex15, getCachePixel((scenePosNew) + faceBitangents[face], face), 0).rgb;
	float luma_c = getLuma(c);
	vec3 d = texelFetch(colortex15, getCachePixel((scenePosNew) + faceTangents[face] + faceBitangents[face], face), 0).rgb;
	float luma_d = getLuma(d);

	float bValid = float(luma_b > 1e-16);
	float cValid = float(luma_c > 1e-16);
	float dValid = float(luma_d > 1e-16);

	vec3 fPos = fract(scenePos + fract(cameraPosition));
	return clamp16F(mix(mix(a, b, fPos[faceHorizontal[face]]*bValid), mix(c, d, fPos[faceHorizontal[face]]*dValid), fPos[faceVertical[face]]*saturate(cValid + dValid)));
}
#endif


vec4 cacheRead(ivec2 storagePos) {
    if (storagePos == ivec2(-1)) return vec4(0.0);
    return clamp16F(vec4(texelFetch(colortex15, storagePos, 0).rgb, 1.0));
}

vec3 cacheReadFog(ivec2 storagePos, vec3 skylight) {
    #ifdef useFogIrradianceVolume
    if (storagePos == ivec2(-1)) return skylight;
    return clamp16F(texelFetch(colortex15, storagePos, 0).rgb);
    #else
    return skylight;
    #endif
}