/* RENDERTARGETS: 15 */
layout(location = 0) out vec4 cacheOut;


#include "/lib/head.glsl"
#include "/lib/util/encoders.glsl"

uniform sampler2D colortex4;
uniform sampler2D colortex5;
uniform sampler2D colortex7;
uniform sampler2D colortex15;

const bool shadowHardwareFiltering = true;

uniform sampler2DShadow shadowtex0;
uniform sampler2DShadow shadowtex1;
uniform sampler2D shadowcolor0;
uniform sampler2D shadowcolor1;

uniform sampler2D noisetex;

uniform int frameCounter;

uniform float frameTime, frameTimeCounter;
uniform float far, near;

uniform float sunAngle;

uniform vec2 viewSize;
uniform vec2 taaOffset;

uniform vec3 cameraPosition, previousCameraPosition;

uniform mat4 gbufferModelView, gbufferModelViewInverse;
uniform mat4 gbufferProjection, gbufferProjectionInverse;
uniform mat4 shadowModelView, shadowModelViewInverse;
uniform mat4 shadowProjection, shadowProjectionInverse;

in vec2 coord;

#define atlasSampler colortex5
#define atlasSampler_s colortex7

#define gFragment

int atlasMipmapLevel = textureQueryLevels(atlasSampler)-1;
ivec2 atlasSize      = ivec2(textureSize(atlasSampler, 0));

#include "/lib/util/colorspace.glsl"
#include "/lib/util/transforms.glsl"

#include "/lib/voxel/store.glsl"
#include "/lib/voxel/intersect.glsl"
#include "/lib/voxel/trace.glsl"

#include "/lib/light/cacheLib.glsl"

struct Material{
    vec3 normal;
    vec3 albedo;
    vec3 emissive;

    int id;
};

#define useMipBounceAlbedo

vec3 getVoxelAlbedo(vec3 pos, vec3 normal, voxelRead voxelData) {
    vec2 uv     = voxelUV(pos, normal, voxelData);

    #ifdef useMipBounceAlbedo
        int lod     = min(min(2, atlasMipmapLevel), 1);
        vec4 albedo = texelFetch(atlasSampler, ivec2(uv * atlasSize / max(lod * 2, 1)), lod);
    #else
        vec4 albedo = texelFetch(atlasSampler, ivec2(uv * atlasSize), 0);
    #endif

    return albedo.rgb * voxelData.tint;
}

void main() {
    ivec2 pixelUV   = ivec2(gl_FragCoord.xy);

    vec4 cacheResult = vec4(0);

    int face    = getFaceFromUV(pixelUV);

    if (face < 7) {
        int framesPerSecond = int(rcp(frameTime));

        vec3 velocity   = vec3(floor(cameraPosition) - floor(previousCameraPosition));
        bool hasMoved   = length(velocity) > 0.2;

        ivec2 facePixel = pixelUV & cacheSegmentEdge;
        vec2 uv         = (vec2(facePixel) + 0.5) / cacheSegmentResolution;
        ivec3 index     = uvToCacheIndex(facePixel);

        vec3 scenePos   = cacheToSceneSpace(index);
        vec3 voxelPos   = sceneToVoxelSpace(scenePos);
        ivec3 voxelIndex = sceneToVoxelSpaceInt(scenePos);

        /*
            Filter method based on ApolloRT by Snurf
        */

        vec4 cacheIn    = texelFetch(colortex15, pixelUV, 0);
        vec3 sum        = cacheIn.rgb;
        vec3 validTaps  = vec3(1.0);

        float blurStrength = saturate(2.0 / max(cacheIn.a, 1.0));

        bool centerOccupied = isVoxelOccupied(voxelIndex, 0);

        if (face < 6 && centerOccupied) {

            blurStrength    = max(blurStrength, 0.01);

            if (!isVoxelOccupied(voxelIndex + ivec3(faceNormals[face]), 0)) {
                for(uint i = 0; i < 6; i++) {
                    vec3 tapLocation = floor(scenePos + vec3(0.5) + faceNormals[i] * 0.6);

                    if (!sceneWithinCacheVolume(tapLocation)) continue;

                    //vec3 tapDirection = normalize(tapLocation - (scenePos + vec3(0.5)));

                    //if (dot(tapDirection, faceNormals[face]) < 0.0) continue;

                    ivec2 UV    = getCachePixel(ivec3(tapLocation), face);

                    if (face != getFaceFromUV(UV)) continue;

                    ivec3 voxelIndex = sceneToVoxelSpaceInt(tapLocation);
                    vec2 voxelCoord = getVoxelCoord(voxelIndex);

                    bool tapOccupied = isVoxelOccupied(voxelIndex, 0);

                    if (!tapOccupied) continue;

                    vec4 tap = texelFetch(colortex15, UV, 0);

                    if (tap.a < 0.5) continue;

                    vec3 tapAlbedo  = vec3(1.0) * blurStrength;

                    if (tapOccupied) {
                        voxelRead currentVoxel = decodeVoxel(uvec4(texture(shadowcolor0, voxelCoord, 0) * 65535.0));

                        tapAlbedo  = getVoxelAlbedo(vec3(voxelIndex) + vec3(0.5), faceNormals[face], currentVoxel) * blurStrength;
                    }

                    sum        += tap.rgb * tapAlbedo;
                    validTaps  += tapAlbedo;
                }
                sum /= max(validTaps, 1e-16);

                cacheResult     = vec4(mix(cacheIn.rgb, sum.rgb, face == 6 ? 0.62 : 0.25), cacheIn.a);
            }
        }
        //cacheResult = cacheIn;

        if (face == 6) {

            blurStrength = max(blurStrength, 0.41);

            for(uint i = 0; i < 6; i++) {
                vec3 tapLocation = floor(scenePos + vec3(0.5) + faceNormals[i] * 0.6);

                if (!sceneWithinCacheVolume(tapLocation)) continue;

                ivec2 UV    = getCachePixel(ivec3(tapLocation), face);

                if (face != getFaceFromUV(UV)) continue;

                ivec3 voxelIndex = sceneToVoxelSpaceInt(tapLocation);
                vec2 voxelCoord = getVoxelCoord(voxelIndex);

                bool tapOccupied = isVoxelOccupied(voxelIndex, 0);

                if (centerOccupied != tapOccupied) continue;

                vec4 tap = texelFetch(colortex15, UV, 0);

                if (tap.a < 0.5) continue;

                vec3 tapAlbedo  = vec3(1.0) * blurStrength;

                if (tapOccupied) {
                    voxelRead currentVoxel = decodeVoxel(uvec4(texture(shadowcolor0, voxelCoord, 0) * 65535.0));

                    tapAlbedo  = getVoxelAlbedo(vec3(voxelIndex) + vec3(0.5), faceNormals[face], currentVoxel) * blurStrength;
                }

                sum        += tap.rgb * tapAlbedo;
                validTaps  += tapAlbedo;
            }
            sum /= max(validTaps, 1e-16);

            cacheResult     = vec4(mix(cacheIn.rgb, sum.rgb, face == 6 ? 0.62 : 0.25), cacheIn.a);

            if (centerOccupied) {
                voxelRead currentVoxel = decodeVoxel(uvec4(texture(shadowcolor0, getVoxelCoord(voxelIndex), 0) * 65535.0));
                if (currentVoxel.ID.x == 0) cacheResult = vec4(0.0);
            }
        }
        //cacheResult = cacheIn;
    }

    cacheOut    = clamp16F(cacheResult);
}