/* RENDERTARGETS: 15 */
layout(location = 0) out vec4 cacheOut;


#include "/lib/head.glsl"
#include "/lib/util/encoders.glsl"

uniform sampler2D colortex4;
uniform sampler2D colortex5;
uniform sampler2D colortex7;
uniform sampler2D colortex15;

const bool shadowHardwareFiltering = true;

uniform sampler2DShadow shadowtex0;
uniform sampler2DShadow shadowtex1;
uniform sampler2D shadowcolor0;
uniform sampler2D shadowcolor1;

uniform sampler2D noisetex;

uniform int frameCounter;

uniform float frameTime, frameTimeCounter;
uniform float far, near;

uniform float sunAngle;

uniform vec2 viewSize;
uniform vec2 taaOffset;
uniform vec2 skyCaptureResolution;

uniform vec3 cameraPosition, previousCameraPosition;
uniform vec3 sunDir, moonDir, lightDir;
uniform vec3 sunDirView, moonDirView, lightDirView;

uniform mat4 gbufferModelView, gbufferModelViewInverse;
uniform mat4 gbufferProjection, gbufferProjectionInverse;
uniform mat4 shadowModelView, shadowModelViewInverse;
uniform mat4 shadowProjection, shadowProjectionInverse;

in vec2 coord;

#ifndef DIM
flat in mat4x3 lightColor;
#elif DIM == -1
flat in mat2x3 lightColor;
#endif

#define atlasSampler colortex5
#define atlasSampler_s colortex7

#define gFragment

int atlasMipmapLevel = textureQueryLevels(atlasSampler)-1;
ivec2 atlasSize      = ivec2(textureSize(atlasSampler, 0));


#include "/lib/atmos/project.glsl"

#include "/lib/util/colorspace.glsl"
#include "/lib/util/transforms.glsl"

#include "/lib/voxel/store.glsl"
#include "/lib/voxel/intersect.glsl"
#include "/lib/voxel/trace.glsl"

#include "/lib/light/cacheLib.glsl"

#include "/lib/frag/bluenoise.glsl"
#include "/lib/frag/gradnoise.glsl"
#include "/lib/mat/emission.glsl"

#include "/lib/brdf/fresnel.glsl"
#include "/lib/brdf/hammon.glsl"

struct Material{
    vec3 normal;
    vec3 albedo;
    vec3 emissive;

    int id;
};

Material hitMaterial(vec3 normal, vec3 albedo, int voxelID) {
    albedo  = linearToAP1Albedo(toLinear(albedo));
    
    return Material(normal, albedo, getEmission(voxelID, albedo), voxelID);
}

float getLabEmission(ivec2 uv) {
    float x     = texelFetch(atlasSampler_s, uv, 0).a * 255;
    return x <= (254.5) ? linStep(x, 0.5, 254.5) : 0.0;
}

vec3 getEmission(int ID, vec3 albedo, ivec2 atlasUV) {
    albedo  = max(albedo, 1e-8);
    vec3 albedoIn   = albedo;

    #if labEmissionMode != 0
    float textureEmission = getLabEmission(atlasUV); 
    #endif

    #if labEmissionMode != 2
        float lum   = getLuma(albedo);

        float albedoLum = mix(avgOf(albedo), maxOf(albedo), 0.71);
            albedoLum   = saturate(albedoLum * sqrt2);

        float emitterLum = saturate(mix(sqr(albedoLum), sqrt(maxOf(albedo)), albedoLum) * sqrt2);

        #if labEmissionMode == 1
            emitterLum  = max(emitterLum, textureEmission);
        #endif

        albedo  = mix(sqr(normalize(albedo)), normalize(albedo), sqrt(emitterLum)) * emitterLum;
    #else       
        albedo   *= textureEmission;
        float emitterLum = textureEmission;
    #endif

    switch (ID) {
        case 1: return albedo * 200;
        case 2: return albedo * 160;
        case 3: return albedo * 80;
        case 10: return LIGHT_TORCH * 120;
        case 11: return LIGHT_REDTORCH * 35;
        case 12: return LIGHT_SOUL * 60;
        case 13: return LIGHT_FIRE * 80;
        case 14: return LIGHT_ENDROD * 100;
    }

    #if labEmissionMode == 0
        return vec3(0);
    #elif labEmissionMode == 1
        return albedoIn * textureEmission * 80;
    #elif labEmissionMode == 2
        return albedo * 80;
    #endif
}

Material hitMaterial(vec3 normal, vec3 albedo, int voxelID, ivec2 atlasUV) {
    albedo  = linearToAP1Albedo(toLinear(albedo));
    
    return Material(normal, albedo, getEmission(voxelID, albedo, atlasUV), voxelID);
}

#define useMipBounceAlbedo

Material getVoxelMat(vec3 pos, vec3 normal, voxelRead voxelData, in int bounceIndex) {
    vec2 uv     = voxelUV(pos, normal, voxelData);

    #ifdef useMipBounceAlbedo
        int lod     = min(min(bounceIndex, atlasMipmapLevel), 1);
        vec4 albedo = texelFetch(atlasSampler, ivec2(uv * atlasSize / max(lod * 2, 1)), lod);
    #else
        vec4 albedo = texelFetch(atlasSampler, ivec2(uv * atlasSize), 0);
    #endif

    return hitMaterial(normal, albedo.rgb * voxelData.tint, voxelData.ID.y, ivec2(uv * atlasSize));
}

#include "/lib/light/warp.glsl"

float readCloudShadowmap(sampler2D shadowmap, vec3 position) {
    position    = mat3(shadowModelView) * position;
    position   /= cloudShadowmapRenderDistance;
    position.xy = position.xy * 0.5 + 0.5;

    position.xy /= vec2(1.0, 1.0 + (1.0 / 3.0));

    return texture(shadowmap, position.xy).a;
}

vec3 getShadowmapPos(vec3 scenePos, const float bias) {  //shadow 2d
    vec3 pos    = scenePos + vec3(bias) * lightDir;
    float a     = length(pos);
        pos     = transMAD(shadowModelView, pos);
        pos     = projMAD(shadowProjection, pos);
        pos.z  *= 0.2;
        pos.z  -= 0.0012*(saturate(a/256.0));

        pos.xy  = shadowmapWarp(pos.xy);

        pos.x   = pos.x * 0.5 + 0.5;
        pos.xy  = pos.yx;

    return pos*0.5+0.5;
}

vec3 getDirectLight(vec3 scenePos, vec3 sceneNormal) {
    float sDotN     = dot(lightDir, sceneNormal);
    if (sDotN <= 0.0) return vec3(0.0);

    float shadow    = 1.0;

    #ifndef DIM
        #ifdef cloudShadowsEnabled
            shadow *= readCloudShadowmap(colortex4, scenePos);

            if (shadow < 1e-4) return vec3(0.0);
        #endif
    #endif

    vec3 shadowPos  = getShadowmapPos(scenePos, 0.08);
        shadow  *= texture(shadowtex1, shadowPos);

    return texture(shadowcolor0, shadowPos.xy).rgb * (shadow * saturate(sDotN));
}


#define HASHSCALE1 .1031
float hash12(vec2 x) {
    vec3 x3  = fract(vec3(x.xyx) * HASHSCALE1);
        x3  += dot(x3, x3.yzx + 19.19);
    return fract((x3.x + x3.y) * x3.z);
}

#define HASHSCALE3 vec3(.1031, .1030, .0973)
vec2 hash22(vec2 p) {
	vec3 p3 = fract(vec3(p.xyx) * HASHSCALE3);
    p3 += dot(p3, p3.yzx + 19.19);
    return fract((p3.xx + p3.yz) * p3.zy);
}
vec2 hash23(vec3 p3) {
    p3 = fract(p3 * HASHSCALE3);
    p3 += dot(p3, p3.yzx + 19.19);
    return fract((p3.xx + p3.yz) * p3.zy);
}

float ditherHashNoise() {
    float noise     = hash12(gl_FragCoord.xy);
        noise   = fract(noise + (frameCounter * 7.0 * rcp(255.0)) *0);

    return noise;
}

float lambertianDiffuse(vec3 normal, vec3 direction) {
    /*
        This is used for some optimized or approximated stuff where Oren-Nayar gives very little visual advantages
        but a big performance loss instead.
    */
    float lambert   = saturate(dot(normal, direction));

    return lambert / pi;
}

uint initRand(uint val0, uint val1, uint backoff) {
    backoff = 16;
	uint v0 = val0, v1 = val1, s0 = 0;

	for (uint n = 0; n < backoff; n++)
	{
		s0 += 0x9e3779b9;
		v0 += ((v1 << 4) + 0xa341316c) ^ (v1 + s0) ^ ((v1 >> 5) + 0xc8013ea4);
		v1 += ((v0 << 4) + 0xad90777d) ^ (v0 + s0) ^ ((v0 >> 5) + 0x7e95761e);
	}
	return v0;
}

// Takes our seed, updates it, and returns a pseudorandom float in [0..1]
float nextRand(inout uint s) {
	s = (1664525u * s + 1013904223u);
	return float(s & 0x00FFFFFF) / float(0x01000000);
}

vec3 genUnitVector(vec2 p) {
    p.x *= tau; p.y = p.y * 2.0 - 1.0;
    return vec3(sincos(p.x) * sqrt(1.0 - p.y * p.y), p.y);
}

vec3 genCosineVector(vec3 vector, vec3 xyz) {
	vec3 dir = genUnitVector(xyz.xy);
	float VoD = dot(vector, dir);

	return sqrt(xyz.z) * (dir - vector * VoD) * inversesqrt(1.0 - VoD * VoD) + sqrt(1.0 - xyz.z) * vector;
}

vec3 GenerateCosineVectorSafe(vec3 vector, vec2 xy) {
	// Apparently this is actually this simple.
	// http://www.amietia.com/lambertnotangent.html
	// (cosine lobe around vector = lambertian BRDF)
	// This one deals with ther rare case where cosineVector == (0,0,0)
	// Can just normalize it instead if you are sure that won't be a problem
	vec3 cosineVector = vector + genUnitVector(xy);
	float lenSq = dot(cosineVector, cosineVector);
	return lenSq > 0.0 ? cosineVector * inversesqrt(lenSq) : vector;
}


uint hashi(uint x, uint y, uint z) {
	x += x >> 11;
	x ^= x << 7;
	x += y;
	x ^= x << 3;
	x += z ^ (x >> 14);
	x ^= x << 6;
	x += x >> 15;
	x ^= x << 5;
	x += x >> 12;
	x ^= x << 9;
	return x;
}

const float seedDelta = 0.001;

vec2 hash2(inout float seed) {
	return fract(sin(vec2(seed+=seedDelta, seed+=seedDelta))*vec2(43758.5453123, 22578.1459123));
}

/**
 * Generate a random value in [-1..+1)
 *   http://amindforeverprogramming.blogspot.de/2013/07/random-floats-in-glsl-330.html
 */
float random(vec2 pos, float time) {
	uint mantissaMask = 0x007FFFFFu;
	uint one = 0x3F800000u;
	uvec3 u = floatBitsToUint(vec3(pos, time));
	uint h = hashi(u.x, u.y, u.z);
	return uintBitsToFloat((h & mantissaMask) | one) - 1.0;
}

#ifdef useEmissionFalloff

#ifndef DIM
    float getEmissionContribution(vec3 voxelPos, vec3 startPos) {
        return cubeSmooth(1.0 / (1.0 + max0(distance(voxelPos, startPos) - 8.0) / 8.0));
    }
#else
    float getEmissionContribution(vec3 voxelPos, vec3 startPos) {
        return cubeSmooth(1.0 / (1.0 + max0(distance(voxelPos, startPos) - 16.0) / 32.0));
    }
#endif

#else
    float getEmissionContribution(vec3 voxelPos, vec3 startPos) {
        return 1.0;
    }
#endif

float volumeContrib(vec3 voxelPos, vec3 startPos) {
    return cubeSmooth(1.0 / (1.0 + max0(distance(voxelPos, startPos) - 16.0) / 16.0));
}

vec3 generateRadiance(vec3 startPos, vec3 direction, const bool volume) {
    vec3 radiance       = vec3(0.0);

    #if DIM != -1
    vec3 directCol      = (sunAngle<0.5 ? lightColor[0] : lightColor[2]) * pi;
    #endif

    voxelRead voxelData;
    vec3 hitNormal;

    vec3 voxelStart = vec3(sceneToVoxelSpaceInt(startPos)) + vec3(0.5);
    vec3 voxelPos   = voxelStart;
    vec3 voxelIndex = floor(voxelPos);

    bool hit        = false;

    vec3 contribution = vec3(1);

    #if OCTREE_LEVELS > 1
    hit     = marchBVH(voxelPos, ivec3(voxelIndex), direction, voxelPos, voxelIndex, hitNormal, voxelData, contribution);
    #else
    hit     = marchVoxel(voxelPos, ivec3(voxelIndex), direction, voxelPos, voxelIndex, hitNormal, voxelData, contribution);
    #endif

    if (hit) {
        vec3 hitPos         = voxelToSceneSpace(voxelPos);
        Material hitMat     =  getVoxelMat(voxelPos, hitNormal, voxelData, 0);
        #if DIM == -1
        hitMat.emissive    *= sqrt2;
        #endif

        radiance           += hitMat.emissive * getEmissionContribution(voxelPos, voxelStart);

        #if DIM != -1
        float sunlightBrdf  = diffuseHammon(hitNormal, direction, lightDir, 1.0);
        radiance           += getDirectLight(hitPos, hitNormal) * sunlightBrdf * directCol * hitMat.albedo;
        #endif

        ivec3 hitIndex      = volume ? 
                    sceneToCacheSpace(floor(hitPos + hitNormal * 0.25)) : 
                    sceneToCacheSpace(floor(hitPos - hitNormal * 0.1));

        int face            = getFace(hitNormal);
        /*
            Propagating from the irradiance volume into the irradiance volume causes bleeding issues.
            Use cache instead.
        */
        #ifndef useIrradianceCache
            if (volume) face = 6;
            radiance           += cacheRead(getCachePixelCacheSpace(hitIndex, face)).rgb * hitMat.albedo * (volume ? volumeContrib(voxelPos, voxelStart) : 1.0);
        #else
            //radiance           += sampleCacheHQ(voxelToSceneSpace(voxelPos), hitNormal, face) * hitMat.albedo;  // interpolation seems to reduce bleeding issues here
            radiance           += cacheRead(getCachePixelCacheSpace(hitIndex, face)).rgb * hitMat.albedo * (volume ? volumeContrib(voxelPos, voxelStart) : 1.0);
        #endif
        
    } else {
        #ifndef DIM
        radiance           += texture(colortex4, projectSky(direction, 2)).rgb * (sstep(direction.y, -0.15, -0.05) * 0.8 + 0.2);
        #endif
    }

    return radiance * contribution;
}

const float piOver4 = 0.785398163397448;
const float piOver2 = 1.570796326794;

vec2 concentricSampleDisk(vec2 random) {
    random = 2.0 * random - 1.0;

    if (all(equal(random, vec2(0)))) return vec2(0);

    float theta, r;
    if (abs(random.x) > abs(random.y)) {
        r = random.x;
        theta = piOver4 * (random.y / random.x);
    } else {
        r = random.y;
        theta = piOver2 - piOver4 * (random.x / random.y);
    }
    return r * vec2(cos(theta), sin(theta));
}
vec3 cosWeightedRandomHemisphereDirection(vec3 n, vec2 r) {
    vec2 d = concentricSampleDisk(r);
    float z = sqrt(max(0.0, 1.0 - d.x * d.x - d.y * d.y));

    vec3  uu = normalize(cross(n, vec3(0.0, 1.0, 1.0)));
    vec3  vv = cross(uu, n);
    vec3  rr = vec3(d.x*uu + d.y*vv + z*n);
    return normalize(rr);
}

void main() {
    ivec2 pixelUV   = ivec2(gl_FragCoord.xy);
    float seed      = random(pixelUV, frameCounter);

    vec4 cacheResult = vec4(0);

    int face    = getFaceFromUV(pixelUV);

    int framesPerSecond = int(rcp(frameTime));

    vec3 velocity           = vec3(floor(cameraPosition) - floor(previousCameraPosition));

    #ifdef useIrradianceCache
    if (face < 6) {
        ivec2 facePixel     = pixelUV & cacheSegmentEdge;
        vec2 uv         = (vec2(facePixel) + 0.5) / cacheSegmentResolution;
        ivec3 index      = uvToCacheIndex(facePixel);

        vec3 scenePos   = cacheToSceneSpace(index);

        bool updateCache = face == (frameCounter % 6);

        if (isVoxelOccupied(sceneToVoxelSpaceInt(scenePos), 0)) {
            vec3 surfacePos = scenePos + vec3(0.5) + faceNormals[face] * 0.51;

            bool isFacingVoxel = isVoxelOccupied(sceneToVoxelSpaceInt(surfacePos), 0);

            if (!isFacingVoxel) {
                vec3 radiance   = vec3(0);

                if(updateCache) {
                    for(uint i = 0; i < cacheSPP; i++) {
                        vec3 rayDir = GenerateCosineVectorSafe(faceNormals[face], hash2(seed));

                        radiance += generateRadiance(surfacePos, rayDir, false);
                    }
                    radiance /= float(cacheSPP);
                }

                vec3 previousScenePos   = (scenePos + velocity);
                ivec2 prevStoragePos    = getCachePixel(previousScenePos, face);

                vec4 history = texelFetch(colortex15, prevStoragePos, 0);

                float samples   = updateCache ? clamp(history.a + 1.0, 1.0, 128.0) : clamp(history.a, 0.0, 128.0);

                radiance = mix(history.rgb, radiance, updateCache ? clamp(exp(-samples), 0.1, 1.0) : 0.0);

                cacheResult     = vec4(radiance, samples);
            }
        }

        //if (updateCache) cacheResult.r = 1.0;
        //else cacheResult.r = 0.0;
    }
    #endif

    #ifdef useFogIrradianceVolume
    if (face == 6) {
        ivec2 facePixel     = pixelUV & cacheSegmentEdge;
        vec2 uv         = (vec2(facePixel) + 0.5) / cacheSegmentResolution;
        ivec3 index      = uvToCacheIndex(facePixel);

        vec3 scenePos   = cacheToSceneSpace(index);

        vec3 surfacePos = scenePos + vec3(0.5);

        ivec3 voxelIndex = sceneToVoxelSpaceInt(surfacePos);

        voxelRead currentVoxel = decodeVoxel(uvec4(texture(shadowcolor0, getVoxelCoord(voxelIndex), 0) * 65535.0));

        bool isBlocked  = isVoxelOccupied(voxelIndex, 0) && currentVoxel.ID.x == 0;

        vec3 radiance   = vec3(0);

        if (!isBlocked) {
            bool updateCache = (frameCounter % 2) == 0;

            if(updateCache) {
                const uint iterations   = radianceVolumeSPP;
                for(uint i = 0; i < iterations; i++) {
                    uint faceIndex = i;
                    vec3 rayDir = genUnitVector(hash2(seed));
                    radiance += generateRadiance(surfacePos, rayDir, true);
                }
                radiance /= float(iterations);
            }

            vec3 previousScenePos   = (scenePos + velocity);
            ivec3 previousIndex     = sceneToCacheSpace(previousScenePos);
            ivec2 prevStoragePos    = getCachePixel(previousScenePos, face);

            vec4 history = texelFetch(colortex15, prevStoragePos, 0);

            float samples   = updateCache || outsideCacheVolume(ivec3(previousIndex)) ? clamp(history.a + 1.0, 1.0, 256.0) : clamp(history.a, 0.0, 256.0);

            radiance = mix(history.rgb, radiance, updateCache ? clamp(exp(-samples), 0.006, 1.0) : 0.0);

            cacheResult     = vec4(radiance, samples);
        } else {
            cacheResult     = vec4(0);
        }
    }
    #endif

    cacheOut    = clamp16F(cacheResult);
}