/*
====================================================================================================

    Copyright (C) 2021 RRe36

    All Rights Reserved unless otherwise explicitly stated.


    By downloading this you have agreed to the license and terms of use.
    These can be found inside the included license-file
    or here: https://rre36.com/copyright-license

    Violating these terms may be penalized with actions according to the Digital Millennium
    Copyright Act (DMCA), the Information Society Directive and/or similar laws
    depending on your country.

====================================================================================================
*/

/* RENDERTARGETS: 5,6,7 */
layout(location = 0) out vec4 GData;
layout(location = 1) out vec4 currentSH;
layout(location = 2) out vec3 chromaData;

#include "/lib/head.glsl"
#include "/lib/util/encoders.glsl"

uniform sampler2D colortex0;
uniform sampler2D colortex1;
uniform sampler2D colortex4;
uniform sampler2D colortex5;
uniform sampler2D colortex7;
uniform sampler2D colortex8;
uniform sampler2D colortex15;

uniform sampler2D depthtex0;

const bool shadowHardwareFiltering = true;

uniform sampler2DShadow shadowtex0;
uniform sampler2DShadow shadowtex1;
uniform sampler2D shadowcolor0;
uniform sampler2D shadowcolor1;

uniform sampler2D noisetex;
uniform sampler3D depthtex2;

uniform int frameCounter;
uniform int frame255;
uniform int frame255Last;

uniform float far, near;

uniform float sunAngle;

uniform vec2 viewSize;
uniform vec2 taaOffset;
uniform vec2 skyCaptureResolution;

uniform vec3 cameraPosition;
uniform vec3 sunDir, moonDir, lightDir;
uniform vec3 sunDirView, moonDirView, lightDirView;

uniform mat4 gbufferModelView, gbufferModelViewInverse;
uniform mat4 gbufferProjection, gbufferProjectionInverse;
uniform mat4 shadowModelView, shadowModelViewInverse;
uniform mat4 shadowProjection, shadowProjectionInverse;

in vec2 coord;

#ifndef DIM
flat in mat4x3 lightColor;
#elif DIM == -1
flat in mat2x3 lightColor;
#endif

#define atlasSampler colortex5
#define atlasSampler_s colortex7

#define gFragment

int atlasMipmapLevel = textureQueryLevels(atlasSampler)-1;
ivec2 atlasSize      = ivec2(textureSize(atlasSampler, 0));

#define GBufferREAD0
#include "/lib/gbuffer.glsl"

#include "/lib/atmos/project.glsl"

#include "/lib/util/colorspace.glsl"
#include "/lib/util/transforms.glsl"

#include "/lib/voxel/store.glsl"
//#include "/lib/voxel/octree.glsl"
#include "/lib/voxel/intersect.glsl"
#include "/lib/voxel/trace.glsl"

#include "/lib/frag/bluenoise.glsl"
#include "/lib/frag/gradnoise.glsl"
#include "/lib/mat/emission.glsl"

#include "/lib/brdf/fresnel.glsl"
#include "/lib/brdf/hammon.glsl"

struct Material{
    vec3 normal;
    vec3 albedo;
    vec3 emissive;

    int id;
};

Material hitMaterial(vec3 normal, vec3 albedo, int voxelID) {
    albedo  = linearToAP1Albedo(toLinear(albedo));
    
    return Material(normal, albedo, getEmission(voxelID, albedo), voxelID);
}

float getLabEmission(ivec2 uv) {
    float x     = texelFetch(atlasSampler_s, uv, 0).a * 255;
    return x <= (254.5) ? linStep(x, 0.5, 254.5) : 0.0;
}

vec3 getEmission(int ID, vec3 albedo, ivec2 atlasUV) {
    albedo  = max(albedo, 1e-8);
    vec3 albedoIn   = albedo;

    #if labEmissionMode != 0
    float textureEmission = getLabEmission(atlasUV); 
    #endif

    #if labEmissionMode != 2
        float lum   = getLuma(albedo);

        float albedoLum = mix(avgOf(albedo), maxOf(albedo), 0.71);
            albedoLum   = saturate(albedoLum * sqrt2);

        float emitterLum = saturate(mix(sqr(albedoLum), sqrt(maxOf(albedo)), albedoLum) * sqrt2);

        #if labEmissionMode == 1
            emitterLum  = max(emitterLum, textureEmission);
        #endif

        albedo  = mix(sqr(normalize(albedo)), normalize(albedo), sqrt(emitterLum)) * emitterLum;
    #else       
        albedo   *= textureEmission;
        float emitterLum = textureEmission;
    #endif

    switch (ID) {
        case 1: return albedo * 200;
        case 2: return albedo * 160;
        case 3: return albedo * 80;
        case 4: return albedo * 10;
        case 10: return LIGHT_TORCH * 120;
        case 11: return LIGHT_REDTORCH * 35;
        case 12: return LIGHT_SOUL * 60;
        case 13: return LIGHT_FIRE * 80;
        case 20: return LIGHT_ENDROD * 100;
        case 21: return LIGHT_AMETHYST * 20;
        case 22: return LIGHT_AMETHYST * 5;
    }

    #if labEmissionMode == 0
        return vec3(0);
    #elif labEmissionMode == 1
        return albedoIn * textureEmission * 80;
    #elif labEmissionMode == 2
        return albedo * 80;
    #endif
}

Material hitMaterial(vec3 normal, vec3 albedo, int voxelID, ivec2 atlasUV) {
    albedo  = linearToAP1Albedo(toLinear(albedo));
    
    return Material(normal, albedo, getEmission(voxelID, albedo, atlasUV), voxelID);
}

#define useMipBounceAlbedo

Material getVoxelMat(vec3 pos, vec3 normal, voxelRead voxelData, in int bounceIndex) {
    vec2 uv     = voxelUV(pos, normal, voxelData);

    #ifdef useMipBounceAlbedo
        int lod     = min(min(bounceIndex, atlasMipmapLevel), 1);
        vec4 albedo = texelFetch(atlasSampler, ivec2(uv * atlasSize / max(lod * 2, 1)), lod);
    #else
        vec4 albedo = texelFetch(atlasSampler, ivec2(uv * atlasSize), 0);
    #endif

    return hitMaterial(normal, albedo.rgb * voxelData.tint, voxelData.ID.y, ivec2(uv * atlasSize));
}

vec2 bluenoiseLookup(ivec2 pos) {
    ivec2 xy    = pos & 255;
    uint z      = frameCounter & 7;

    return texelFetch(depthtex2, ivec3(xy, z), 0).xy;
}

vec2 bluenoiseLookup(ivec2 pos, int zIndex) {
    ivec2 xy    = pos & 255;
    uint z      = zIndex & 7;

    return texelFetch(depthtex2, ivec3(xy, z), 0).xy;
}

vec2 bluenoiseLookupGrad(ivec2 pos) {
    ivec2 xy    = (pos + frameCounter) & 255;
    uint z      = frameCounter & 7;

    return texelFetch(depthtex2, ivec3(xy, z), 0).xy;
}

vec2 unprojectSphere(vec3 dir) {
    vec2 lonlat     = vec2(atan(-dir.x, dir.z), acos(dir.y));
    return lonlat * vec2(rcp(tau), rcp(pi)) + vec2(0.5, 0.0);
}

float depthLinear(float depth) {
    return (2.0*near) / (far+near-depth * (far-near));
}

#include "/lib/light/warp.glsl"

float readCloudShadowmap(sampler2D shadowmap, vec3 position) {
    position    = mat3(shadowModelView) * position;
    position   /= cloudShadowmapRenderDistance;
    position.xy = position.xy * 0.5 + 0.5;

    position.xy /= vec2(1.0, 1.0 + (1.0 / 3.0));

    return texture(shadowmap, position.xy).a;
}

vec3 getShadowmapPos(vec3 scenePos, const float bias) {  //shadow 2d
    vec3 pos    = scenePos + vec3(bias) * lightDir;
    float a     = length(pos);
        pos     = transMAD(shadowModelView, pos);
        pos     = projMAD(shadowProjection, pos);
        pos.z  *= 0.2;
        pos.z  -= 0.0012*(saturate(a/256.0));

        pos.xy  = shadowmapWarp(pos.xy);

        pos.x   = pos.x * 0.5 + 0.5;
        pos.xy  = pos.yx;

    return pos*0.5+0.5;
}

vec3 getDirectLight(vec3 scenePos, vec3 sceneNormal) {
    float sDotN     = dot(lightDir, sceneNormal);
    if (sDotN <= 0.0) return vec3(0.0);

    float shadow    = 1.0;

    #ifndef DIM
        #ifdef cloudShadowsEnabled
            shadow *= readCloudShadowmap(colortex4, scenePos);

            if (shadow < 1e-4) return vec3(0.0);
        #endif
    #endif

    vec3 shadowPos  = getShadowmapPos(scenePos, 0.08);
        shadow  *= texture(shadowtex1, shadowPos);

    return texture(shadowcolor0, shadowPos.xy).rgb * (shadow * saturate(sDotN));
}


#define HASHSCALE1 .1031
float hash12(vec2 x) {
    vec3 x3  = fract(vec3(x.xyx) * HASHSCALE1);
        x3  += dot(x3, x3.yzx + 19.19);
    return fract((x3.x + x3.y) * x3.z);
}

#define HASHSCALE3 vec3(.1031, .1030, .0973)
vec2 hash22(vec2 p) {
	vec3 p3 = fract(vec3(p.xyx) * HASHSCALE3);
    p3 += dot(p3, p3.yzx + 19.19);
    return fract((p3.xx + p3.yz) * p3.zy);
}
vec2 hash23(vec3 p3) {
    p3 = fract(p3 * HASHSCALE3);
    p3 += dot(p3, p3.yzx + 19.19);
    return fract((p3.xx + p3.yz) * p3.zy);
}

float ditherHashNoise() {
    float noise     = hash12(gl_FragCoord.xy);
        noise   = fract(noise + (frameCounter * 7.0 * rcp(255.0)) *0);

    return noise;
}

#define rtHashNoise

float lambertianDiffuse(vec3 normal, vec3 direction) {
    /*
        This is used for some optimized or approximated stuff where Oren-Nayar gives very little visual advantages
        but a big performance loss instead.
    */
    float lambert   = saturate(dot(normal, direction));

    return lambert / pi;
}

uint initRand(uint val0, uint val1, uint backoff) {
    backoff = 16;
	uint v0 = val0, v1 = val1, s0 = 0;

	for (uint n = 0; n < backoff; n++)
	{
		s0 += 0x9e3779b9;
		v0 += ((v1 << 4) + 0xa341316c) ^ (v1 + s0) ^ ((v1 >> 5) + 0xc8013ea4);
		v1 += ((v0 << 4) + 0xad90777d) ^ (v0 + s0) ^ ((v0 >> 5) + 0x7e95761e);
	}
	return v0;
}

// Takes our seed, updates it, and returns a pseudorandom float in [0..1]
float nextRand(inout uint s) {
	s = (1664525u * s + 1013904223u);
	return float(s & 0x00FFFFFF) / float(0x01000000);
}

vec3 genUnitVector(vec2 p) {
    p.x *= tau; p.y = p.y * 2.0 - 1.0;
    return vec3(sincos(p.x) * sqrt(1.0 - p.y * p.y), p.y);
}

vec3 genCosineVector(vec3 vector, vec3 xyz) {
	vec3 dir = genUnitVector(xyz.xy);
	float VoD = dot(vector, dir);

	return sqrt(xyz.z) * (dir - vector * VoD) * inversesqrt(1.0 - VoD * VoD) + sqrt(1.0 - xyz.z) * vector;
}

vec3 GenerateCosineVectorSafe(vec3 vector, vec2 xy) {
	// Apparently this is actually this simple.
	// http://www.amietia.com/lambertnotangent.html
	// (cosine lobe around vector = lambertian BRDF)
	// This one deals with ther rare case where cosineVector == (0,0,0)
	// Can just normalize it instead if you are sure that won't be a problem
	vec3 cosineVector = vector + genUnitVector(xy);
	float lenSq = dot(cosineVector, cosineVector);
	return lenSq > 0.0 ? cosineVector * inversesqrt(lenSq) : vector;
}


/*
    Based on ApolloRT by Snurf
*/

const float piOver4 = 0.785398163397448;
const float piOver2 = 1.570796326794;

vec2 concentricSampleDisk(vec2 random) {
    random = 2.0 * random - 1.0;

    if (all(equal(random, vec2(0)))) return vec2(0);

    float theta, r;
    if (abs(random.x) > abs(random.y)) {
        r = random.x;
        theta = piOver4 * (random.y / random.x);
    } else {
        r = random.y;
        theta = piOver2 - piOver4 * (random.x / random.y);
    }
    return r * vec2(cos(theta), sin(theta));
}
vec3 cosWeightedRandomHemisphereDirection(vec3 n, vec2 r) {
    vec2 d = concentricSampleDisk(r);
    float z = sqrt(max(0.0, 1.0 - d.x * d.x - d.y * d.y));

    vec3  uu = normalize(cross(n, vec3(0.0, 1.0, 1.0)));
    vec3  vv = cross(uu, n);
    vec3  rr = vec3(d.x*uu + d.y*vv + z*n);
    return normalize(rr);
}

vec3 screenspaceRT(vec3 position, vec3 direction, float noise) {
    const uint maxSteps     = 10;
    const float stepSize    = halfPi / float(maxSteps);

    vec3 stepVector         = direction * stepSize;

    vec3 endPosition        = position + stepVector * maxSteps;
    vec3 endScreenPosition  = viewToScreenSpace(endPosition);

    vec2 maxPosXY           = max(abs(endScreenPosition.xy * 2.0 - 1.0), vec2(1.0));
    float stepMult          = minOf(vec2(1.0) / maxPosXY);
        stepVector         *= stepMult;

    // closest texel iteration
    vec3 samplePos          = position;
    
        samplePos          += stepVector / 6.0;
    vec3 screenPos          = viewToScreenSpace(samplePos);

    const float thresh      = 0.25;

    if (saturate(screenPos.xy) == screenPos.xy) {
        float depthSample   = texelFetch(depthtex0, ivec2(screenPos.xy * viewSize), 0).x;
        float linearSample  = depthLinear(depthSample);
        float currentDepth  = depthLinear(screenPos.z);

        if (linearSample < currentDepth) {
            float dist      = abs(linearSample - currentDepth) / clamp(currentDepth, 0.25 / far, 2.0 / far);
            if (dist <= thresh) return vec3(screenPos.xy, depthSample);
        }
    }
    
        samplePos          += stepVector * noise;

    for (uint i = 0; i < maxSteps; ++i) {
        vec3 screenPos      = viewToScreenSpace(samplePos);
            samplePos      += stepVector;
        if (saturate(screenPos.xy) != screenPos.xy) break;

        float depthSample   = texelFetch(depthtex0, ivec2(screenPos.xy * viewSize), 0).x;
        float linearSample  = depthLinear(depthSample);
        float currentDepth  = depthLinear(screenPos.z);

        if (linearSample < currentDepth) {
            float dist      = abs(linearSample - currentDepth) / clamp(currentDepth, 0.25 / far, 2.0 / far);
            if (dist <= thresh) return vec3(screenPos.xy, depthSample);
        }
    }

    return vec3(1.1);
}

mat2x3 unpackDoubleRGBE8(vec4 data){
    vec3 Color_A    = decodeRGBE8(vec4(unpack2x8(data.x), unpack2x8(data.y)));
    vec3 Color_B    = decodeRGBE8(vec4(unpack2x8(data.z), unpack2x8(data.w)));

    return mat2x3(Color_A, Color_B);
}

bool screenRaytrace(vec3 pos, vec3 direction, inout vec3 hitPosition, inout vec3 hitIndex, inout vec3 hitNormal, inout Material hitMat, inout vec3 light) {
        pos     = voxelToSceneSpace(pos);
        pos     = viewToSceneSpace(pos, gbufferModelView);

        direction   = mat3(gbufferModelView) * direction;
    
    vec3 hitPos     = screenspaceRT(pos, direction, ditherHashNoise());

    if (hitPos.z >= 1.0) return false;

    ivec2 pixelPos  = ivec2(hitPos.xy * viewSize);

    vec4 tex1       = texelFetch(colortex1, pixelPos, 0);

    mat2x3 hitColor = unpackDoubleRGBE8(texelFetch(colortex8, pixelPos, 0));

    vec3 albedo     = texelFetch(colortex0, pixelPos, 0).rgb;

        hitNormal   = decodeNormal(tex1.xy);

    hitMat      = Material(hitNormal, albedo, hitColor[1], 0);
    light       = hitColor[0] * albedo;

    hitPos      = screenToViewSpace(hitPos, true);
    hitPos      = viewToSceneSpace(hitPos);

    hitPosition = sceneToVoxelSpace(hitPos);
    hitIndex    = floor(hitPosition - hitNormal / 128.0);

    return true;
}

#ifdef useEmissionFalloff

#ifndef DIM
    float getEmissionContribution(vec3 voxelPos, vec3 startPos) {
        return cubeSmooth(1.0 / (1.0 + max0(distance(voxelPos, startPos) - 8.0) / 8.0));
    }
#else
    float getEmissionContribution(vec3 voxelPos, vec3 startPos) {
        return cubeSmooth(1.0 / (1.0 + max0(distance(voxelPos, startPos) - 16.0) / 32.0));
    }
#endif

#else
    float getEmissionContribution(vec3 voxelPos, vec3 startPos) {
        return 1.0;
    }
#endif

#include "/lib/light/sh.glsl"

#include "/lib/light/cacheLib.glsl"

SphericalHarmonics PTLight(vec3 scenePos, vec3 sceneNormals, Material surfaceMat, uint randSeed, vec2 uv) {
    const float a1 = 1.0 / rho;
    const float a2 = a1 * a1;

    voxelRead voxelData;

    vec2 quasirandomCurr2 = 0.5 + fract(vec2(a1, a2) * frame255 + 0.5);
    
    vec3 startPos       = sceneToVoxelSpace(scenePos);

    #if DIM != -1
    vec3 directCol      = (sunAngle<0.5 ? lightColor[0] : lightColor[2]) * pi;

    mat2x3 lighting     = unpackDoubleRGBE8(texelFetch(colortex8, ivec2(uv * viewSize), 0));
    bool inSunlight     = avgOf(lighting[0]) > rpi;
    #endif

    SphericalHarmonics totalColor       = InitSH();

    for (int rayIndex = 0; rayIndex < PT_SPP; ++rayIndex) {
        int frame255new     = frame255 + int(rayIndex * 256.0);

        vec2 randSinCos     = vec2(cos(quasirandomCurr2.x), sin(quasirandomCurr2.y));

        vec2 noiseCurr  = hash22(gl_FragCoord.xy + frame255new);

        vec2 quasirandomCurr = 0.5 + fract(vec2(a1, a2) * frame255new + 0.5);

        vec3 contribution = vec3(1.0);

        Material hitMat = surfaceMat;

        vec3 hitNormal = hitMat.normal;

        vec3 voxelPos   = sceneToVoxelSpace(scenePos);
        vec3 voxelIndex = floor(voxelPos - sceneNormals / 128.0);
        vec3 startIndex = floor(voxelPos - sceneNormals / 128.0);
        vec3 cameraPos  = sceneToVoxelSpace(gbufferModelViewInverse[3].xyz);

        vec3 direction  = normalize(voxelPos - cameraPos);

        bool hitSky     = false;

        //++quasirandomCurr2;

        mat2x3 rayColor = mat2x3(0.0);

        vec3 firstBounceDir     = vec3(0);

        bool doSSPT     = true;

        #ifdef useIrradianceCache
        for (int i = 0; i < 1; ++i) {
        #else
        for (int i = 0; i < bounceNumber; ++i) {
            #if DIM != -1
            if (i >= 2 && inSunlight) break;
            #endif
        #endif

            ++quasirandomCurr;

            noiseCurr      += hash22(vec2(gl_FragCoord.xy + vec2(cos(quasirandomCurr.x), sin(quasirandomCurr.y))));

            bool initialVoxel   = false;
            vec3 oldDirection   = direction;

            vec2 vectorXY   = fract(sqrt(2.0) * quasirandomCurr + noiseCurr);

            direction       = GenerateCosineVectorSafe(hitNormal, vectorXY);
            //direction       = cosWeightedRandomHemisphereDirection(hitNormal, vectorXY);

            if (i == 0) firstBounceDir = direction;

            vec3 brdf = hitMat.albedo;
                brdf *= diffuseHammon(hitNormal, -oldDirection, direction, 1.0) / saturate(dot(direction, hitNormal)/pi);

            contribution *= saturate(brdf);

            if(dot(hitNormal, direction) < 0.0 && !initialVoxel) continue;

            bool hit;

            #ifdef SSPT_Enabled
                if (doSSPT) {
                    vec3 directLight = vec3(0.0);
                        hit    = screenRaytrace(voxelPos, direction, voxelPos, voxelIndex, hitNormal, hitMat, directLight);

                    if (hit) {
                        #ifdef stainedGlassTintingEnabled
                        ivec3 neighbourIndex = ivec3(floor(voxelPos + hitNormal / 64.0));
                        if (isVoxelOccupied(neighbourIndex, 0)) {
                            voxelRead voxel = decodeVoxel(uvec4(texture(shadowcolor0, getVoxelCoord(neighbourIndex), 0) * 65535.0));
                            if (voxel.ID.x == 50) {
                                contribution *= voxelAbsorption(voxelPos, hitNormal, voxel, atlasSampler);
                            }
                        }
                        #endif

                        rayColor[0] += contribution * hitMat.emissive;
                        #if DIM != -1
                        rayColor[1] += contribution * directCol * directLight;
                        #endif

                        hitSky      = false;

                        continue;
                    } else {
                        doSSPT      = false;
                    }
                }
            #endif

            #if OCTREE_LEVELS > 1
            hit     = marchBVH(voxelPos, ivec3(voxelIndex), direction, voxelPos, voxelIndex, hitNormal, voxelData, contribution);
            #else
            hit     = marchVoxel(voxelPos, ivec3(voxelIndex), direction, voxelPos, voxelIndex, hitNormal, voxelData, contribution);
            #endif

            if (hit) {
                hitMat      = getVoxelMat(voxelPos, hitNormal, voxelData, i);

                rayColor[0] += contribution * hitMat.emissive * getEmissionContribution(voxelPos, startPos);

                #if DIM != -1
                rayColor[1] += contribution * directCol * getDirectLight(voxelToSceneSpace(voxelPos), hitNormal) * hitMat.albedo;
                #endif

                hitSky      = false;
            }

            if (!hit) {
                hitSky      = true;

                break;
            }

            if (distance(voxelPos, startPos) > maxTraversalDistance) break;
        }

        if (hitSky) {
            #ifndef DIM
            rayColor[1]   += contribution * texture(colortex4, projectSky(direction, 2)).rgb * (sstep(direction.y, -0.15, -0.05) * 0.8 + 0.2);
            #elif DIM == -1
            rayColor[1]   += contribution * lightColor[0] / halfPi;
            #endif
        }

        #ifdef useIrradianceCache
		int face = getFace(hitNormal);
		//vec3 indirect = sampleCacheHQ(voxelToSceneSpace(voxelPos), hitNormal, face) * hitMat.albedo * contribution;
        vec3 indirect = cacheRead(getCachePixel(voxelToSceneSpace(voxelPos) - hitNormal * 0.05, face)).rgb * hitMat.albedo * contribution;
        #else
        const vec3 indirect     = vec3(0.0);
        #endif

        #if DIM == -1
        rayColor[0]    *= sqrt2;
        #endif

        //rayColor[1]    *= sqrt2;

        accumulateSH(totalColor, IrradianceToSH((rayColor[0] + rayColor[1]) + indirect, firstBounceDir), 1.0 / PT_SPP);
    }
    //rayColor[0]     = max0(rayColor[0] / PT_SPP);
    //rayColor[1]     = max0(rayColor[1] / PT_SPP) * sqrt2;

    //divideSH(totalColor, PT_SPP);

    return totalColor;
}

ivec2 clampTexelPos(ivec2 pos) {
    return clamp(pos, ivec2(0.0), ivec2(viewSize));
}



void main() {
    vec2 lowresCoord    = coord / indirectResScale;
    ivec2 pixelPos      = ivec2(floor(coord * viewSize - taaOffset * 0.5) / indirectResScale);
    float sceneDepth    = texelFetch(depthtex0, pixelPos, 0).x;

    currentSH       = vec4(0);
    chromaData      = vec3(0);
    GData           = vec4(0.0, 0.0, 1.0, 1.0);

    uint randSeed = initRand(uint(gl_FragCoord.x + gl_FragCoord.y * viewSize.x), uint(frameCounter), 16);

    if (landMask(sceneDepth) && saturate(lowresCoord) == lowresCoord) {
        vec2 coord      = saturate(lowresCoord);

        ivec2 UV        = ivec2(coord * viewSize);
        GData00 Gbuffer0 = ReadGbuffer0(UV);

        vec3 viewPos    = screenToViewSpace(vec3(coord, sceneDepth), true);
        vec3 scenePos   = viewToSceneSpace(viewPos);

        GData           = saturate(vec4(Gbuffer0.SceneNormal * 0.5 + 0.5, sqrt(depthLinear(sceneDepth))));

        Material surfaceMat = Material(Gbuffer0.SceneNormal, vec3(1.0), vec3(0.0), Gbuffer0.BlockID.y);

        SphericalHarmonics result = PTLight(scenePos, Gbuffer0.SceneNormal, surfaceMat, randSeed, lowresCoord);

        currentSH       = result.Y;
        chromaData.xy   = result.CoCg;
        chromaData.z    = pow5(Gbuffer0.Lightmap.x);
        //currentFrame    = clamp16F(result[0] + result[1]);
    }

    currentSH   = clamp16FN(currentSH);
    chromaData  = clamp16FN(chromaData);
}