/*
====================================================================================================

    Copyright (C) 2021 RRe36

    All Rights Reserved unless otherwise explicitly stated.


    By downloading this you have agreed to the license and terms of use.
    These can be found inside the included license-file
    or here: https://rre36.com/copyright-license

    Violating these terms may be penalized with actions according to the Digital Millennium
    Copyright Act (DMCA), the Information Society Directive and/or similar laws
    depending on your country.

====================================================================================================
*/

/* RENDERTARGETS: 6,7 */
layout(location = 0) out vec4 resultSH;
layout(location = 1) out vec4 resultAux;

#include "/lib/head.glsl"
#include "/lib/util/encoders.glsl"

in vec2 coord;

uniform sampler3D colortex4;
uniform sampler2D colortex5;
uniform sampler2D colortex6;
uniform sampler2D colortex7;

uniform int frameCounter;

uniform float far;

uniform vec2 pixelSize, viewSize;

/* ------ ATROUS ------ */

ivec2 clampUV_LR(ivec2 UV) {
    return clamp(UV, ivec2(0.0), ivec2(viewSize * indirectResScale - 1));
}
bool UV_WithinBounds(ivec2 UV) {
    return clamp(UV, ivec2(0), ivec2(viewSize * indirectResScale - 1)) == UV;
}

const float gaussKernel[2][2] = {
	{ 1.0 / 4.0, 1.0 / 8.0  },
	{ 1.0 / 8.0, 1.0 / 16.0 }
};

float computeSigmaL(ivec2 pixelUV, float center) {
	float sum = center * gaussKernel[0][0];

    const int r = 1;
	for(int y = -r; y <= r; y++) {
		for(int x = -r; x <= r; x++) {
			if(x != 0 || y != 0) {
				ivec2 UV    = pixelUV + ivec2(x, y);
				float variance = texelFetch(colortex7, UV, 0).z;
				float w     = gaussKernel[abs(x)][abs(y)];
				sum        += variance * w;
			}
		}
	}

	return sqrt(max(sum, 1e-8));
}

#include "/lib/light/sh.glsl"

#define gbufferImage colortex5

vec4 getGbufferData(ivec2 uv) {
    vec4 g  = texelFetch(gbufferImage, uv, 0);
        g.w *= g.w;
    return (g * vec4(vec3(2.0), far)) - vec4(vec3(1.0), 0.0);
}

vec4 getGbufferData(vec2 uv) {
    vec4 g  = textureLod(gbufferImage, uv, 0);
        g.w *= g.w;
    return (g * vec4(vec3(2.0), far)) - vec4(vec3(1.0), 0.0);
}

vec2 temporalBluenoise() {
    ivec3 uv    = ivec3(ivec2(gl_FragCoord.xy) & 255, frameCounter & 7);

    return texelFetch(colortex4, uv, 0).xy;
}

SphericalHarmonics filterImage(out float varOut, vec2 uv, const int stepSize) {
	ivec2 pixelUV       = ivec2(uv * viewSize);

    vec4 normalDepthCenter  = getGbufferData(pixelUV);

    vec4 colorCenterY   = texelFetch(colortex6, pixelUV, 0);
    vec4 colorCenterAux = texelFetch(colortex7, pixelUV, 0);

    SphericalHarmonics Color = SphericalHarmonics(colorCenterY, colorCenterAux.xy);

    if (normalDepthCenter.a == far) return Color;

	float lumaCenter = getLuma(ToIrradiance(Color));

    float frames    = colorCenterAux.w * maxFrames;

    float normalExp = mix(2.0, 8.0, saturate(frames / 8.0)) * SVGF_NORMALEXP;

    //float sigmaBias = (1.0 / max(frames, 1.0)) + 0.1;
    float sigmaBias = exp(-frames * PT_SPP / 8.0) + 0.16;
    float maxDelta  = mix(halfPi, tau, saturate(frames / 32.0));
    float offset    = mix(0.04 / (0.5 * SVGF_RAD), 0.01, saturate(frames / 32.0));
    float sigmaMul  = mix(1.0, 0.15, saturate(frames / 16.0));

    float sigmaDistMul = 2.0 - (1.0 / (1.0 + normalDepthCenter.a / 128.0));

	float sigmaL = 1.0 / (sigmaMul * SVGF_STRICTNESS * sigmaDistMul * computeSigmaL(pixelUV, colorCenterAux.z) + sigmaBias * sigmaDistMul);

	float sumVariance = colorCenterAux.z;
    float sumWeight = 1.0;

    ivec2 jitter        = ivec2(temporalBluenoise() - 0.5) * stepSize;

	const int r = SVGF_RAD;
	for(int y = -r; y <= r; ++y) {
		for(int x = -r; x <= r; ++x) {
			ivec2 p = pixelUV + ivec2(x, y) * stepSize + jitter;

			if(x == 0 && y == 0)
				continue;

			float w = float(all(greaterThanEqual(p, ivec2(0)))
					&& all(lessThan(p, ivec2(viewSize * indirectResScale))));

            vec4 shAux      = texelFetch(colortex7, p, 0);
			SphericalHarmonics col = SphericalHarmonics(texelFetch(colortex6, p, 0), shAux.xy);

			float lum       = getLuma(ToIrradiance(col));

			float distLum   = abs(lumaCenter - lum);
                distLum     = sqr(distLum) / max(lumaCenter, offset);
                distLum     = clamp(distLum, 0.0, maxDelta);

            vec4 normalDepth    = getGbufferData(p);
            if (normalDepth.a == far) continue;

			float distDepth = abs(normalDepthCenter.a - normalDepth.a) * 4.0;

			w *= pow(max(0.0, dot(normalDepthCenter.rgb, normalDepth.rgb)), normalExp);
			w *= exp(-distDepth / sqrt(float(stepSize)) - sqrt(distLum * sigmaL));

            accumulateSH(Color, col, w);

			sumVariance    += shAux.z * sqr(w);
			sumWeight      += w;
		}
	}
    divideSH(Color, sumWeight);

	sumVariance /= sqr(sumWeight);

    varOut      = sumVariance;

	return Color;
}

void main() {
    resultSH        = texelFetch(colortex6, ivec2(coord * viewSize), 0);
    resultAux       = texelFetch(colortex7, ivec2(coord * viewSize), 0);

    vec2 lowresCoord    = coord / indirectResScale;
    if (saturate(lowresCoord) != lowresCoord) discard;

    #ifdef SVGF_POSTACCUM
    SphericalHarmonics result = filterImage(resultAux.z, coord, SVGF_SIZE);

    resultSH        = result.Y;
    resultAux.xy    = result.CoCg;
    #endif

    resultSH        = clamp16FN(resultSH);
    resultAux       = clamp16FN(resultAux);
}