#include "/lib/head.glsl"
#include "/lib/util/encoders.glsl"

layout (triangles) in;

#if (defined useInstancing || !defined generateShadowmap)
layout (triangle_strip, max_vertices = 3) out;
#else
layout (triangle_strip, max_vertices = 6) out;
#endif

in vertexOut {
    flat int InstanceID;
    flat int voxelMode;
    flat int blockID;

    flat ivec2 objectID;

    vec2 UV;
    vec2 MidUV;

    vec3 Tint;
    vec3 LocalLocation;
    vec3 SceneLocation;
    vec3 CenteredSceneLocation;
    flat vec3 Normal;

    flat vec4 Tangent;
} vertexIn[3];

out voxelGeometryOut {
    flat int isVoxel;
    flat uvec4 voxelWrite;
};

out shadowGeometryOut {
    flat int matID;
    float warp;
    vec2 uv;
    vec3 tint;
    vec3 scenePos;
    vec3 worldPos;
};

uniform int instanceId;

uniform ivec2 atlasSize;

uniform vec3 cameraPosition;

uniform mat4 shadowModelViewInverse;

#include "/lib/light/warp.glsl"

#include "/lib/voxel/store.glsl"

struct voxelData {
    ivec2 voxelID;
    vec3 tint;
};
struct atlasData {
    vec2 tileIndexFraction; //2x8 bits
    vec2 tileIndexFloored;  //2x4 bits
    float tileRes;          //4 bits
};

uvec4 packVoxelData0(voxelData data, atlasData atlas) {
    uvec4 bf;
        bf.x    = pack2x8UI(data.voxelID);
        bf.y    = packColor16UI(data.tint);
        bf.z    = pack2x8UI(atlas.tileIndexFraction);
        bf.w    = pack2x8UI(vec2(encode2x4Unorm(atlas.tileIndexFloored), atlas.tileRes));

    return bf;
}

void main() {
    #ifdef useInstancing
    if (vertexIn[0].InstanceID == 0) {
    #endif

    #ifdef generateShadowmap
    for (int i = 0; i < 3; ++i) {
        isVoxel = 0;
        tint    = vertexIn[i].Tint;
        uv      = vertexIn[i].UV;
        matID   = vertexIn[i].blockID;

        gl_Position     = gl_ModelViewMatrix * vec4(vertexIn[i].LocalLocation, 1.0);

        scenePos        = vertexIn[i].SceneLocation;
        worldPos        = cameraPosition + scenePos;

        gl_Position     = gl_ProjectionMatrix * gl_Position;
        gl_Position.xy  = shadowmapWarp(gl_Position.xy, warp);
        gl_Position.z  *= 0.2;
        gl_Position.x   = gl_Position.x * 0.5 + 0.5;
        gl_Position.xy  = gl_Position.yx;
        if (gl_Position.y <= 0.0) return;

        EmitVertex();
    } EndPrimitive();
    #endif

    #ifdef useInstancing
        return;
    } else {
    #endif

    if (vertexIn[0].voxelMode == 0) return;

    vec3 bitangent  = cross(vertexIn[0].Tangent.xyz, vertexIn[0].Normal) * (vertexIn[0].Tangent.w < 0.0 ? -1.0 : 1.0);
    vec3 vertPosU   = vertexIn[0].Tangent.xyz * mat3(vertexIn[0].LocalLocation, vertexIn[1].LocalLocation, vertexIn[2].LocalLocation);
    vec3 vertPosV   = bitangent * mat3(vertexIn[0].LocalLocation, vertexIn[1].LocalLocation, vertexIn[2].LocalLocation);

    vec2 faceSize   = vec2(maxOf(vertPosU) - minOf(vertPosU), maxOf(vertPosV) - minOf(vertPosV));
        faceSize   /= vec2(length(vertexIn[0].Tangent.xyz), length(bitangent));

    vec2 minCoord   = min(min(vertexIn[0].UV, vertexIn[1].UV), vertexIn[2].UV);
    vec2 maxCoord   = max(max(vertexIn[0].UV, vertexIn[1].UV), vertexIn[2].UV);
    vec2 tileSize   = abs(maxCoord - minCoord) / faceSize;
    vec2 tileRes    = round(tileSize * atlasSize);

    ivec2 tileIndex = ivec2(vertexIn[0].MidUV / tileSize);

    voxelData voxelTemp     = voxelData(vertexIn[0].objectID, vertexIn[0].Tint);
    atlasData atlasTemp     = atlasData(fract(tileIndex / 256.0) * 256.0 / 255.0,
                                        floor(tileIndex / 256.0) / 15.0,
                                        log2(maxOf(tileRes)) / 15.0);

    uvec4 voxelWriteTemp    = packVoxelData0(voxelTemp, atlasTemp);

    vec3 normal     = mat3(shadowModelViewInverse) * normalize(gl_NormalMatrix * vertexIn[0].Normal);

    float normalBias = 0.5 - 0.25 * normal.y;

    vec3 triangleBoundsMin  = min(vertexIn[0].CenteredSceneLocation, min(vertexIn[1].CenteredSceneLocation, vertexIn[2].CenteredSceneLocation));
    vec3 triangleBoundsMax  = max(vertexIn[0].CenteredSceneLocation, max(vertexIn[1].CenteredSceneLocation, vertexIn[2].CenteredSceneLocation));

    vec3 triangleMidLocation = (triangleBoundsMin + triangleBoundsMax) * 0.5;

    vec3 voxelIndex     = floor(sceneToVoxelSpace(triangleMidLocation));

    if (withinVoxelVolume(voxelIndex)) {
        vec2 texelCoord     = getVoxelCoord(voxelIndex) * 2.0 - 1.0;

        const vec2[3] verts = vec2[3](vec2(1,-1) / shadowMapRes, vec2(0,1) / shadowMapRes, vec2(-1,-1) / shadowMapRes);

        for (int j = 0; j < 3; ++j) {
            voxelWrite      = voxelWriteTemp;
            isVoxel         = 1;

            gl_Position     = vec4(texelCoord, normalBias, 1.0);
            gl_Position.xy += verts[j];

            EmitVertex();
        }
        EndPrimitive();
    } else {
        return;
    }

    #ifdef useInstancing
        return;
    }
    #endif
}
