#version 450 compatibility

/*
====================================================================================================

    Copyright (C) 2020 RRe36

    All Rights Reserved unless otherwise explicitly stated.


    By downloading this you have agreed to the license and terms of use.
    These can be found inside the included license-file
    or here: https://rre36.com/copyright-license

    Violating these terms may be penalized with actions according to the Digital Millennium
    Copyright Act (DMCA), the Information Society Directive and/or similar laws
    depending on your country.

====================================================================================================
*/

/* RENDERTARGETS: 0,3 */
layout(location = 0) out vec3 sceneColor;
layout(location = 1) out vec4 bloomTiles;

#include "/lib/head.glsl"
#include "/lib/util/encoders.glsl"

in vec2 coord;

flat in mat4x3 lightColor;

uniform sampler2D colortex0;
uniform sampler2D colortex1;
uniform sampler2D colortex2;
uniform sampler2D colortex4;
uniform sampler2D colortex5;
uniform sampler2D colortex6;
uniform sampler2D colortex7;
uniform sampler2D colortex8;
uniform sampler2D colortex15;

uniform sampler2D depthtex0;
uniform sampler2D depthtex1;

uniform sampler2D noisetex;

uniform int frameCounter;
uniform int isEyeInWater;
uniform int worldTime;

uniform float eyeAltitude;
uniform float far, near;
uniform float frameTimeCounter;
uniform float lightFlip;
uniform float sunAngle;
uniform float wetness;

uniform vec2 skyCaptureResolution;
uniform vec2 pixelSize, viewSize;
uniform vec2 taaOffset;

uniform vec3 cameraPosition;
uniform vec3 lightDir, lightDirView;

uniform vec4 daytime;

uniform ivec2 eyeBrightness;
uniform ivec2 eyeBrightnessSmooth;

uniform mat4 gbufferModelView, gbufferModelViewInverse;
uniform mat4 gbufferProjection, gbufferProjectionInverse;

#define GBufferREAD0
#define GBufferREAD1
#include "/lib/gbuffer.glsl"

#define FUTIL_LINDEPTH
#include "/lib/fUtil.glsl"
#include "/lib/frag/bluenoise.glsl"
#include "/lib/frag/gradnoise.glsl"
#include "/lib/util/transforms.glsl"

#include "/lib/atmos/waterConst.glsl"

#include "/lib/atmos/phase.glsl"
#include "/lib/atmos/air/const.glsl"
#include "/lib/atmos/project.glsl"

#include "/lib/atmos/fog.glsl"

void applyFogData(inout vec3 color, in mat2x3 data) {
    color = color * data[1] + data[0];
}

mat2x3 unpackDoubleRGBE8(vec4 data){
    vec3 Color_A    = decodeRGBE8(vec4(unpack2x8(data.x), unpack2x8(data.y)));
    vec3 Color_B    = decodeRGBE8(vec4(unpack2x8(data.z), unpack2x8(data.w)));

    return mat2x3(Color_A, Color_B);
}

/* REFLECTIONS */
#include "/lib/util/colorspace.glsl"
#include "/lib/brdf/fresnel.glsl"
#include "/lib/brdf/material.glsl"
#include "/lib/brdf/specular.glsl"
#include "/lib/frag/capture.glsl"

/*
    These two functions used for rough reflections are based on zombye's spectrum shaders
    https://github.com/zombye/spectrum
*/

mat3 getRotationMat(vec3 x, vec3 y) {
	float cosine = dot(x, y);
	vec3 axis = cross(y, x);

	float tmp = 1.0 / dot(axis, axis);
	      tmp = tmp - tmp * cosine;
	vec3 tmpv = axis * tmp;

	return mat3(
		axis.x * tmpv.x + cosine, axis.x * tmpv.y - axis.z, axis.x * tmpv.z + axis.y,
		axis.y * tmpv.x + axis.z, axis.y * tmpv.y + cosine, axis.y * tmpv.z - axis.x,
		axis.z * tmpv.x - axis.y, axis.z * tmpv.y + axis.x, axis.z * tmpv.z + cosine
	);
}
vec3 ggxFacetDist(vec3 viewDir, float roughness, vec2 xy) {
	/*
        GGX VNDF sampling
        http://www.jcgt.org/published/0007/04/01/
    */

    viewDir     = normalize(vec3(roughness * viewDir.xy, viewDir.z));

    float clsq  = dot(viewDir.xy, viewDir.xy);
    vec3 T1     = vec3(clsq > 0.0 ? vec2(-viewDir.y, viewDir.x) * inversesqrt(clsq) : vec2(1.0, 0.0), 0.0);
    vec3 T2     = vec3(-T1.y * viewDir.z, viewDir.z * T1.x, viewDir.x * T1.y - T1.x * viewDir.y);

	float r     = sqrt(xy.x);
	float phi   = tau * xy.y;
	float t1    = r * cos(phi);
	float a     = saturate(1.0 - t1 * t1);
	float t2    = mix(sqrt(a), r * sin(phi), 0.5 + 0.5 * viewDir.z);

	vec3 normalH = t1 * T1 + t2 * T2 + sqrt(saturate(a - t2 * t2)) * viewDir;

	return normalize(vec3(roughness * normalH.xy, normalH.z));
}

mat2x3 unpackReflectionAux(vec4 data){
    vec3 shadows    = decodeRGBE8(vec4(unpack2x8(data.x), unpack2x8(data.y)));
    vec3 albedo     = decodeRGBE8(vec4(unpack2x8(data.z), unpack2x8(data.w)));

    return mat2x3(shadows, albedo);
}

vec3 screenspaceRT(vec3 position, vec3 direction, float noise) {
    const uint maxSteps     = 16;

  	float rayLength = ((position.z + direction.z * far * sqrt3) > -near) ?
                      (-near - position.z) / direction.z : far * sqrt3;

    vec3 screenPosition     = viewToScreenSpace(position);
    vec3 endPosition        = position + direction * rayLength;
    vec3 endScreenPosition  = viewToScreenSpace(endPosition);

    vec3 screenDirection    = normalize(endScreenPosition - screenPosition);
        screenDirection.xy  = normalize(screenDirection.xy);

    vec3 maxLength          = (step(0.0, screenDirection) - screenPosition) / screenDirection;
    float stepMult          = minOf(maxLength);
    vec3 screenVector       = screenDirection * stepMult / float(maxSteps);

    vec3 screenPos          = screenPosition + screenDirection * maxOf(pixelSize * pi);

    if (saturate(screenPos.xy) == screenPos.xy) {
        float depthSample   = texelFetch(depthtex0, ivec2(screenPos.xy * viewSize), 0).x;
        float linearSample  = depthLinear(depthSample);
        float currentDepth  = depthLinear(screenPos.z);

        if (linearSample < currentDepth) {
            float dist      = abs(linearSample - currentDepth) / currentDepth;
            if (dist <= 0.25) return vec3(screenPos.xy, depthSample);
        }
    }

        screenPos          += screenVector * noise;

    for (uint i = 0; i < maxSteps; ++i) {
        if (saturate(screenPos.xy) != screenPos.xy) break;

        float depthSample   = texelFetch(depthtex0, ivec2(screenPos.xy * viewSize), 0).x;
        float linearSample  = depthLinear(depthSample);
        float currentDepth  = depthLinear(screenPos.z);

        if (linearSample < currentDepth) {
            float dist      = abs(linearSample - currentDepth) / currentDepth;
            if (dist <= 0.5) return vec3(screenPos.xy, depthSample);
        }

        screenPos      += screenVector;
    }

    return vec3(1.1);
}

vec3 screenspaceRT_LR(vec3 position, vec3 direction, float noise) {
    const uint maxSteps     = 8;

  	float rayLength = ((position.z + direction.z * far * sqrt3) > -near) ?
                      (-near - position.z) / direction.z : far * sqrt3;

    vec3 screenPosition     = viewToScreenSpace(position);
    vec3 endPosition        = position + direction * rayLength;
    vec3 endScreenPosition  = viewToScreenSpace(endPosition);

    vec3 screenDirection    = normalize(endScreenPosition - screenPosition);
        screenDirection.xy  = normalize(screenDirection.xy);

    vec3 maxLength          = (step(0.0, screenDirection) - screenPosition) / screenDirection;
    float stepMult          = min(minOf(maxLength), 0.5);
    vec3 screenVector       = screenDirection * stepMult / float(maxSteps);

    vec3 screenPos          = screenPosition;

        screenPos          += screenVector * noise;

    for (uint i = 0; i < maxSteps; ++i) {
        if (saturate(screenPos.xy) != screenPos.xy) break;

        float depthSample   = texelFetch(depthtex0, ivec2(screenPos.xy * viewSize), 0).x;
        float linearSample  = depthLinear(depthSample);
        float currentDepth  = depthLinear(screenPos.z);

        if (linearSample < currentDepth) {
            float dist      = abs(linearSample - currentDepth) / currentDepth;
            if (dist <= 0.5) return vec3(screenPos.xy, depthSample);
        }

        screenPos      += screenVector;
    }

    return vec3(1.1);
}

vec4 readSkyCapture(vec3 direction, float occlusion) {
    return vec4(texture(colortex4, projectSky(direction, 2)).rgb * sqrt(occlusion), sqrt(occlusion));
}

const bool shadowHardwareFiltering = true;

uniform sampler2DShadow shadowtex0;
uniform sampler2DShadow shadowtex1;
uniform sampler2D shadowcolor0;
uniform sampler2D shadowcolor1;

uniform mat4 shadowModelView, shadowModelViewInverse;
uniform mat4 shadowProjection, shadowProjectionInverse;

#include "/lib/light/warp.glsl"

float readCloudShadowmap(sampler2D shadowmap, vec3 position) {
    position    = mat3(shadowModelView) * position;
    position   /= cloudShadowmapRenderDistance;
    position.xy = position.xy * 0.5 + 0.5;

    position.xy /= vec2(1.0, 1.0 + (1.0 / 3.0));

    return texture(shadowmap, position.xy).a;
}

vec3 getShadowmapPos(vec3 scenePos, const float bias) {  //shadow 2d
    vec3 pos    = scenePos + vec3(bias) * lightDir;
    float a     = length(pos);
        pos     = transMAD(shadowModelView, pos);
        pos     = projMAD(shadowProjection, pos);
        pos.z  *= 0.2;
        pos.z  -= 0.0012*(saturate(a/256.0));

        pos.xy  = shadowmapWarp(pos.xy);

        pos.x   = pos.x * 0.5 + 0.5;
        pos.xy  = pos.yx;

    return pos*0.5+0.5;
}
vec3 getDirectLight(vec3 scenePos, vec3 sceneNormal) {
    float sDotN     = dot(lightDir, sceneNormal);
    if (sDotN <= 0.0) return vec3(0.0);

    float shadow    = 1.0;

    #ifndef DIM
        #ifdef cloudShadowsEnabled
            shadow *= readCloudShadowmap(colortex4, scenePos);

            if (shadow < 1e-4) return vec3(0.0);
        #endif
    #endif

    vec3 shadowPos  = getShadowmapPos(scenePos, 0.08);
        shadow  *= texture(shadowtex1, shadowPos);

    return texture(shadowcolor0, shadowPos.xy).rgb * (shadow * saturate(sDotN));
}

#define atlasSampler colortex8
//#define atlasSampler_s colortex7

int atlasMipmapLevel = textureQueryLevels(atlasSampler)-1;
ivec2 atlasSize      = ivec2(textureSize(atlasSampler, 0));

#define RPASS

#include "/lib/voxel/store.glsl"
//#include "/lib/voxel/octree.glsl"
#include "/lib/voxel/intersect.glsl"
#include "/lib/voxel/trace.glsl"
#include "/lib/mat/emission.glsl"

struct Material{
    vec3 normal;
    vec3 albedo;
    vec3 emissive;

    int id;
};

Material hitMaterial(vec3 normal, vec3 albedo, int voxelID) {
    albedo  = linearToAP1Albedo(toLinear(albedo));
    
    return Material(normal, albedo, getEmission(voxelID, albedo), voxelID);
}

float getLabEmission(ivec2 uv) {
    /*
    float x     = texelFetch(atlasSampler_s, uv, 0).a * 255;
    return x <= (254.5) ? linStep(x, 0.5, 254.5) : 0.0;*/
    return 0.0;
}

vec3 getEmission(int ID, vec3 albedo, ivec2 atlasUV) {
    albedo  = max(albedo, 1e-8);
    vec3 albedoIn   = albedo;

    #if labEmissionMode != 0
    float textureEmission = getLabEmission(atlasUV); 
    #endif

    #if labEmissionMode != 2
        float lum   = getLuma(albedo);

        float albedoLum = mix(avgOf(albedo), maxOf(albedo), 0.71);
            albedoLum   = saturate(albedoLum * sqrt2);

        float emitterLum = saturate(mix(sqr(albedoLum), sqrt(maxOf(albedo)), albedoLum) * sqrt2);

        #if labEmissionMode == 1
            emitterLum  = max(emitterLum, textureEmission);
        #endif

        albedo  = mix(sqr(normalize(albedo)), normalize(albedo), sqrt(emitterLum)) * emitterLum;
    #else       
        albedo   *= textureEmission;
        float emitterLum = textureEmission;
    #endif

    switch (ID) {
        case 1: return albedo * 200;
        case 2: return albedo * 160;
        case 3: return albedo * 80;
        case 4: return albedo * 10;
        case 10: return LIGHT_TORCH * 120;
        case 11: return LIGHT_REDTORCH * 35;
        case 12: return LIGHT_SOUL * 60;
        case 13: return LIGHT_FIRE * 80;
        case 20: return LIGHT_ENDROD * 100;
        case 21: return LIGHT_AMETHYST * 20;
        case 22: return LIGHT_AMETHYST * 5;
    }

    #if labEmissionMode == 0
        return vec3(0);
    #elif labEmissionMode == 1
        return albedoIn * textureEmission * 80;
    #elif labEmissionMode == 2
        return albedo * 80;
    #endif
}

Material hitMaterial(vec3 normal, vec3 albedo, int voxelID, ivec2 atlasUV) {
    albedo  = linearToAP1Albedo(toLinear(albedo));
    
    return Material(normal, albedo, getEmission(voxelID, albedo, atlasUV), voxelID);
}

#define useMipBounceAlbedo

Material getVoxelMat(vec3 pos, vec3 normal, voxelRead voxelData, in int bounceIndex) {
    vec2 uv     = voxelUV(pos, normal, voxelData);

    #ifdef useMipBounceAlbedo
        int lod     = min(min(bounceIndex, atlasMipmapLevel), 1);
        vec4 albedo = texelFetch(atlasSampler, ivec2(uv * atlasSize / max(lod * 2, 1)), lod);
    #else
        vec4 albedo = texelFetch(atlasSampler, ivec2(uv * atlasSize), 0);
    #endif

    return hitMaterial(normal, albedo.rgb * voxelData.tint, voxelData.ID.y, ivec2(uv * atlasSize));
}

#include "/lib/light/cacheLib.glsl"
#include "/lib/brdf/hammon.glsl"

#define HASHSCALE3 vec3(.1031, .1030, .0973)
vec2 hash22(vec2 p) {
	vec3 p3 = fract(vec3(p.xyx) * HASHSCALE3);
    p3 += dot(p3, p3.yzx + 19.19);
    return fract((p3.xx + p3.yz) * p3.zy);
}
vec3 genUnitVector(vec2 p) {
    p.x *= tau; p.y = p.y * 2.0 - 1.0;
    return vec3(sincos(p.x) * sqrt(1.0 - p.y * p.y), p.y);
}
vec3 GenerateCosineVectorSafe(vec3 vector, vec2 xy) {
	// Apparently this is actually this simple.
	// http://www.amietia.com/lambertnotangent.html
	// (cosine lobe around vector = lambertian BRDF)
	// This one deals with ther rare case where cosineVector == (0,0,0)
	// Can just normalize it instead if you are sure that won't be a problem
	vec3 cosineVector = vector + genUnitVector(xy);
	float lenSq = dot(cosineVector, cosineVector);
	return lenSq > 0.0 ? cosineVector * inversesqrt(lenSq) : vector;
}

vec4 reflectionTrace(vec3 viewPosition, vec3 viewNormal, vec3 viewDirection) {
    vec3 scenePosition  = viewToSceneSpace(viewPosition);
    vec3 sceneNormal    = mat3(gbufferModelViewInverse) * viewNormal;
    vec3 direction      = mat3(gbufferModelViewInverse) * viewDirection;

    vec3 voxelPos       = sceneToVoxelSpace(scenePosition);
    vec3 voxelIndex     = floor(voxelPos - sceneNormal / 128.0);
    vec3 startIndex     = voxelIndex;
    vec3 cameraPosition = sceneToVoxelSpace(gbufferModelViewInverse[3].xyz);

    vec3 rayColor       = vec3(0);
    vec3 contribution   = vec3(1);

    const float a1 = 1.0 / rho;
    const float a2 = a1 * a1;

    voxelRead voxelData;
    Material hitMat;

    vec3 directCol      = (sunAngle<0.5 ? lightColor[0] : lightColor[2]) * pi;

    vec3 hitNormal      = sceneNormal;

    bool hit            = marchBVH(voxelPos, ivec3(voxelIndex), direction, voxelPos, voxelIndex, hitNormal, voxelData, contribution);

    if (hit) {
        hitMat      = getVoxelMat(voxelPos, hitNormal, voxelData, 0);
        rayColor   += contribution * (hitMat.emissive + (directCol * getDirectLight(voxelToSceneSpace(voxelPos), hitNormal)) * hitMat.albedo);

        vec2 quasirandomCurr = 0.5 + fract(vec2(a1, a2));
        vec2 noiseCurr  = hash22(gl_FragCoord.xy);

        noiseCurr      += hash22(vec2(gl_FragCoord.xy + vec2(cos(quasirandomCurr.x), sin(quasirandomCurr.y))));
        vec3 oldDirection   = direction;

        vec2 vectorXY   = fract(sqrt(2.0) * quasirandomCurr + noiseCurr);

        direction       = GenerateCosineVectorSafe(hitNormal, vectorXY);

        vec3 brdf = hitMat.albedo;
            brdf *= diffuseHammon(hitNormal, -oldDirection, direction, 1.0) / saturate(dot(direction, hitNormal)/pi);

        contribution *= saturate(brdf);

        hit = marchBVH(voxelPos, ivec3(voxelIndex), direction, voxelPos, voxelIndex, hitNormal, voxelData, contribution);
        int face = getFace(hitNormal);
        if (hit) rayColor   += contribution * (hitMat.emissive + (directCol * getDirectLight(voxelToSceneSpace(voxelPos), hitNormal) + cacheRead(getCachePixel(voxelToSceneSpace(voxelPos) - hitNormal * 0.05, face)).rgb) * hitMat.albedo);

    return vec4(rayColor, 1.0);
    }

    return vec4(0);
}

void main() {
    sceneColor = stex(colortex0).rgb;

    vec2 sceneDepth = vec2(stex(depthtex0).x, stex(depthtex1).x);

    vec3 viewPos0   = screenToViewSpace(vec3(coord, sceneDepth.x));
    vec3 scenePos0  = viewToSceneSpace(viewPos0);

    vec3 viewPos1   = screenToViewSpace(vec3(coord, sceneDepth.y));
    vec3 scenePos1  = viewToSceneSpace(viewPos1);

    vec3 viewDir    = normalize(viewPos0);
    vec3 worldDir   = normalize(scenePos0);

    float vDotL     = dot(normalize(viewPos1), lightDirView);

    float caveMult  = sqr(linStep(eyeBrightnessSmooth.y/240.0, 0.1, 0.9));

    float bluenoise     = ditherBluenoise();

    vec3 fogcol     = (sunAngle<0.5 ? lightColor[0] : lightColor[1]) * lightFlip * mieHG(vDotL, 0.68) * sqr(caveMult);
        fogcol     *= mix(0.33, 1.0, sqrt(abs(lightDir.y)));
        fogcol     += lightColor[2] * caveMult * rcp(pi);

    if (landMask(sceneDepth.x)) {
        ivec2 UV        = ivec2(coord * viewSize);

        GData00 Gbuffer0 = ReadGbuffer0(UV);
        GData01 Gbuffer1 = ReadGbuffer1(UV);

        vec3 viewNormal = mat3(gbufferModelView) * Gbuffer0.SceneNormal;

        bool water      = Gbuffer0.BlockID.x == 102;

        mat2x3 lightingData = unpackDoubleRGBE8(texelFetch(colortex5, ivec2(coord * viewSize), 0));

        vec3 viewDir    = normalize(viewPos0);

        mat2x3 reflectionAux = lightingData;
            reflectionAux[0] *= pi;

        vec3 directCol  = (sunAngle<0.5 ? lightColor[0] : lightColor[2]) * lightFlip;
            reflectionAux[0] = min(reflectionAux[0], directCol);

        materialProperties material = decodeLabBasic(Gbuffer1.MaterialData.xy);
        if (water) material = materialProperties(0.0001, 0.02, false, false, mat2x3(0.0));

        if (dot(viewDir, viewNormal) > 0.0) viewNormal = -viewNormal;

        vec3 reflectDir = reflect(viewDir, viewNormal);
        
        vec4 reflection = vec4(0.0);
        vec3 fresnel    = vec3(0.0);

        float skyOcclusion  = cubeSmooth(sqr(linStep(Gbuffer0.Lightmap.y, skyOcclusionThreshold - 0.2, skyOcclusionThreshold)));

        #ifdef resourcepackReflectionsEnabled
            /* --- ROUGH REFLECTIONS --- */

            float roughnessFalloff  = 1.0 - (linStep(material.roughness, roughnessThreshold * 0.71, roughnessThreshold));

            #ifdef roughReflectionsEnabled
            if (material.roughness < 0.0002 || water) {
            #endif

                vec3 reflectSceneDir = mat3(gbufferModelViewInverse) * reflectDir;

                #ifdef screenspaceReflectionsEnabled
                    vec3 reflectedPos = screenspaceRT(viewPos0, reflectDir, ditherBluenoise());
                    if (reflectedPos.z < 1.0) reflection += vec4(texelFetch(colortex0, ivec2(reflectedPos.xy * viewSize), 0).rgb, 1.0);
                    else {
                        vec4 voxelSample = reflectionTrace(viewPos0, viewNormal, reflectDir);
                        if (voxelSample.a > 0.0) reflection += vec4(voxelSample.rgb, 1.0);
                        else reflection += readSkyCapture(reflectSceneDir, 1.0);
                    }
                #else
                    reflection += readSkyCapture(reflectSceneDir, skyOcclusion);
                #endif

                    if (clamp16F(reflection) != reflection) reflection = vec4(0.0);

                    fresnel    += BRDFfresnelAlbedoTint(-viewDir, viewNormal, material, reflectionAux[1]);

            #ifdef roughReflectionsEnabled
            
            } else {
                mat3 rot        = getRotationMat(vec3(0, 0, 1), viewNormal);
                vec3 tangentV   = viewDir * rot;
                float noise     = ditherBluenoise();
                float dither    = ditherGradNoiseTemporal();

                const uint steps    = roughReflectionSamples;
                const float rSteps  = 1.0 / float(steps);

                for (uint i = 0; i < steps; ++i) {
                    if (roughnessFalloff <= 1e-3) break;
                    vec2 xy         = vec2(fract((i + noise) * sqr(32.0) * phi), (i + noise) * rSteps);
                    vec3 roughNrm   = rot * ggxFacetDist(-tangentV, material.roughness, xy);

                    vec3 reflectDir = reflect(viewDir, roughNrm);

                    vec3 reflectSceneDir = mat3(gbufferModelViewInverse) * reflectDir;

                    #ifdef screenspaceReflectionsEnabled
                        vec3 reflectedPos       = vec3(1.1);
                        if (material.roughness < 0.3) reflectedPos = screenspaceRT(viewPos0, reflectDir, dither);
                        else if (material.roughness < 0.7) reflectedPos = screenspaceRT_LR(viewPos0, reflectDir, dither);

                        if (reflectedPos.z < 1.0) reflection += vec4(texelFetch(colortex0, ivec2(reflectedPos.xy * viewSize), 0).rgb, 1.0);
                        else {
                            vec4 voxelSample = reflectionTrace(viewPos0, roughNrm, reflectDir);
                            if (voxelSample.a > 0.0) reflection += vec4(voxelSample.rgb, 1.0);
                            else reflection += readSkyCapture(reflectSceneDir, 1.0);
                        }
                    #else
                        reflection += readSkyCapture(reflectSceneDir, skyOcclusion);
                    #endif

                        fresnel    += BRDFfresnelAlbedoTint(-viewDir, roughNrm, material, reflectionAux[1]);
                }
                if (clamp16F(reflection) != reflection) reflection = vec4(0.0);

                reflection *= rSteps;
                fresnel    *= rSteps;

                reflection.a *= roughnessFalloff;
            }

            #else
                reflection.a *= roughnessFalloff;
            #endif

            if (material.conductor) {
                float lumratio  = getLuma(reflection.rgb) / max(getLuma(sceneColor.rgb), 1e-8);
                reflection.rgb += vec3(getLuma(sceneColor.rgb)) * (1.0 - sstep(lumratio, 0.1, 0.3)) * rpi * (1.0 - sstep(getLuma(reflection.rgb), 0.01, 0.1));
            }

            if (material.conductor) sceneColor.rgb = mix(sceneColor.rgb, reflection.rgb * fresnel, reflection.a);
            else sceneColor.rgb = mix(sceneColor.rgb, reflection.rgb, fresnel * reflection.a);            
        #else
            /* --- WATER REFLECTIONS --- */
            if (water) {
                vec3 reflectSceneDir = mat3(gbufferModelViewInverse) * reflectDir;

                vec3 reflectedPos = screenspaceRT(viewPos0, reflectDir, ditherBluenoise());
                if (reflectedPos.z < 1.0) reflection += vec4(texelFetch(colortex0, ivec2(reflectedPos.xy * viewSize), 0).rgb, 1.0);
                else reflection += readSkyCapture(reflectSceneDir, skyOcclusion);

                if (clamp16F(reflection) != reflection) reflection = vec4(0.0);

                    fresnel    += BRDFfresnel(-viewDir, viewNormal, material, reflectionAux[1]);
                sceneColor.rgb = mix(sceneColor.rgb, reflection.rgb, fresnel * reflection.a);
            }
        #endif

        if (water) sceneColor.rgb += specularBeckmann(-viewDir, lightDirView, viewNormal, material) * reflectionAux[0];
        #ifdef specularHighlightsEnabled
        else sceneColor.rgb += specularTrowbridgeReitzGGX(-viewDir, lightDirView, viewNormal, material, reflectionAux[1]) * reflectionAux[0];
        #endif
    }

    #ifndef waterVolumeEnabled
        if (isEyeInWater==1) sceneColor.rgb = waterFog(sceneColor.rgb, length(scenePos0), fogcol);
    #endif

    #ifdef fogVolumeEnabled
        mat2x3 fogData    = mat2x3(stex(colortex6).rgb, stex(colortex7).rgb);
        applyFogData(sceneColor.rgb, fogData);
    #endif

    sceneColor          = drawbufferClamp(sceneColor);
    bloomTiles          = vec4(0);
}